/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
#ifndef GUARD_connector_h
#define GUARD_connector_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file connector.h
/// @brief Class for network connector header file.

#include <fstream>

#include <ace/Reactor.h>
#include <ace/INET_Addr.h>
#include <ace/SOCK_Stream.h>
#include <ace/SOCK_Connector.h>
#include <ace/Connector.h>
#include <ace/Svc_Handler.h>
#include <ace/Reactor_Notification_Strategy.h>

#include <glibmm.h>

namespace NetUtil {

    enum HTTPHeaderStatus {

        UNKNOWN_ERROR = 0,
        INFORMATIONAL = 1,
        SUCCESS = 2,
        REDIRECTION = 3,
        CLIENT_ERROR = 4,
        SERVER_ERROR = 5

    };

    void SendGetMessage(const ACE_TCHAR* request,
                        ACE_SOCK_Stream& peer,
                        const ACE_TCHAR* accept = "*/*");

    void SendGetMessage(const ACE_TCHAR* request,
                        const ACE_Byte* author,
                        ACE_SOCK_Stream& peer,
                        const ACE_TCHAR* accept = "*/*");
};

/// @brief Class for HTTP URL parsing.
class Address
{
public:
    Address(const ACE_TCHAR* str);
    ~Address();

    /// @brief Return host name.
    const char* RawHostName() const
    {
        return rawHostName_;
    }

    /// @brief Return request.
    const char* Request() const
    {
        return request_;
    }

    /// @brief Return full address.
    ///
    /// Return Orignal string except author string.
    const char* FullAddress() const
    {
        return fullAddress_;
    }

    /// @brief Return BASE64 encoded author string.
    const ACE_Byte* Authorization() const
    {
        return encodedAuthorizationString_;
    }

    /// @brief Return port number.
    const u_short& PortNumber() const
    {
        return portNumber_;
    }

private:
    static const char* httpProtocolString_;
    static const size_t httpProtocolStringLength_ = 7;

    u_short portNumber_;

    char* rawHostName_;
    char* request_;

    char* fullAddress_;

    char* authorizationString_;

    ACE_Byte* encodedAuthorizationString_;

    void SplitAddressString(const ACE_TCHAR* str);
    void GetAuthorizationString(const ACE_TCHAR* str);
    void GetRawHostname(const ACE_TCHAR* str);
    void GetRequest(const ACE_TCHAR* str);
    void GetFullAddress(const ACE_TCHAR* str);

    void EncodeAuthorizationString();

    void UrlEncode(const unsigned char* orig, unsigned char* dest);

    /// @brief Convert decimal to hex.
    ///
    /// If orig is less than 9, return 0~9(ascii 48~57).
    /// If orig is over than 9, return A~F(ascii 65~70).
    unsigned char DecToHex(const unsigned char orig)
    {
        if(orig < 10)
            return orig + 48;
        else
            return orig + 55;
    }

    /// @brief If orig need urlencode, return true.
    /// 
    /// According to RFC1738, check is char need to urlencode.
    /// This function does not check slash char in reserved characters.
    bool NeedUrlEncode(unsigned char orig)
    {
        if(orig > 122 ||
           orig < 33 ||
           orig > 57 && orig < 65 ||
           orig > 90 && orig < 95 ||
           orig > 33 && orig < 39 ||
           orig == 43 ||
           orig == 44 ||
           //orig == 47 ||
           orig == 96)
            return true;
        else
            return false;
    }
};

std::auto_ptr<Address> CreateAddress(const std::string& destUrl,
                                     bool convertCurrentEncoding = false);

/// @brief Class for get file from http server.
class FileGet : public ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_MT_SYNCH>
{
    typedef ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_MT_SYNCH> FileGetHandler;

public:
    FileGet():
        recvHeader_(false),
        redirect_(false),
        notifier_(0, this, ACE_Event_Handler::WRITE_MASK),
        destAddress_(NULL),
        outfile_(NULL),
        redirectURL_(NULL)
    {
        ACE_OS::memset(headerBuffer_, 0, sizeof(headerBuffer_));
    }

    ~FileGet()
    {
        if(redirectURL_ != NULL)
            delete [] redirectURL_;

        if(outfile_ != NULL)
            ACE_OS::fclose(outfile_);
    }

    /// @brief Set destination address.
    void SetDestAddress(Address* destAddress)
    {
        destAddress_ = destAddress;
    }

    /// @brief Set output file.
    void SetOutFile(const ACE_TCHAR* filename)
    {
        outfile_ = ACE_OS::fopen(filename, "w");
    }

    /// @brief Receive redirect or not.
    bool Redirect()
    {
        return redirect_;
    }

    /// @brief Return redirect url.
    const char* RedirectURL()
    {
        return redirectURL_;
    }

    virtual int open(void *p = 0);

    virtual int handle_input(ACE_HANDLE fd = ACE_INVALID_HANDLE);
    virtual int handle_output(ACE_HANDLE fd = ACE_INVALID_HANDLE);
    virtual int handle_timeout(const ACE_Time_Value& currentTime,
                               const void* act = 0);
private:
    enum {
        ITERATIONS = 5
    };

    bool recvHeader_;
    bool redirect_;

    int iterations_;
    ACE_Reactor_Notification_Strategy notifier_;

    Address* destAddress_;

    FILE* outfile_;
    char* redirectURL_;

    char headerBuffer_[1024];
};

#endif
