/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
#ifndef GUARD_dirwnd_h
#define GUARD_dirwnd_h
// Sagua is music player.
//
// Orignal widget first appear in 'Air' project(http://air.kldp.net),
// and this is compact type. 
//
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file dirbrowsewin.h
/// @brief Class for directory browse window header file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if HAVE_EXT_HASH_SET && __GNUC__ >= 3
#include <ext/hash_set>
#else
#include <set>
#endif

#include <string>

#include <gdkmm.h>

#include <gtkmm/window.h>
#include <gtkmm/treemodel.h>
#include <gtkmm/treestore.h>
#include <gtkmm/treeview.h>
#include <gtkmm/menu.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/separator.h>
#include <gtkmm/buttonbox.h>
#include <gtkmm/button.h>
#include <gtkmm/stock.h>
#include <gtkmm/dialog.h>

/// @brief Class for directory browse window.
///
/// To check existence of destination path in treestore, 
/// need external container. If you can use hase set, try hash set to 
/// external container(not standard STL yet). OTherwise use std::set.
///
/// Directory add to tree process.
///
/// 1. Add a root directory(/) at first.
///
/// 2. Add a dummy directory under root.
///
/// 3. When expander is opened, read directories and add them.
///
/// 4. If something is appended, add a dummy directory under that.
class DirBrowseWnd : public Gtk::ScrolledWindow
{
public:
    DirBrowseWnd(const int& border = 12, const bool& displayWatch = true, const std::string& initialDirectory = "");
    virtual ~DirBrowseWnd();

    /// @brief Get selected directory full path.
    const std::string& SelectedDirectory() const { return selectedDirectory_; }
    const std::string& SelectedDirectoryFromIter(); 

    /// @brief Get signal handler for directory select signal.
    sigc::signal<void> SignalSelected() { return signalSelected_; }

private:
#if HAVE_EXT_HASH_SET && __GNUC__ >= 3
    struct StringHash
    {
        size_t operator()(const std::string& id) const
        {
            const std::string::size_type size(id.size());
            size_t h(0);
            for(std::string::size_type i(0); i < size; ++i) {

		    h = 5 * h + id[i];
            }

            return h;
        }
    };
	
    __gnu_cxx::hash_set<std::string, StringHash> pathInStore_;

    void ResetPathSet()
    {
        __gnu_cxx::hash_set<std::string, StringHash>().swap(pathInStore_);
    }

#else
    std::set<std::string> pathInStore_;

    void ResetPathSet()
    {
        std::set<std::string>().swap(pathInStore_);
    }
	
#endif
    /// @brief Class for column used in DirBrowseWnd
    class ModelColumns : public Gtk::TreeModel::ColumnRecord
    {
    public:
        ModelColumns() { add(colName_); add(colPath_); }

        // colName_ is display element, so i'll try it Glib::ustring.
        Gtk::TreeModelColumn<Glib::ustring> colName_;	///< Short path name to display.
        Gtk::TreeModelColumn<std::string> colPath_;	///< Full path name. 
    };

    ModelColumns column_;			    ///< Column used in DirBrowseWnd.	
    
    Gtk::TreeView treeView_;		    ///< Treeview to display column.

    Gtk::Menu menuPopup_;			    ///< Menu to use refresh menu.
	
    static std::string selectedDirectory_;	    ///< Selected directory.

    std::string currentPath_;		    ///< Temporary string used in OnRowExpanded.
    std::string fullPath_;			    ///< Temporary string used in OnRowExpanded.

    Glib::RefPtr<Gtk::TreeStore> treeModel_;

    sigc::signal<void> signalSelected_;	    ///< Signal handler for column select signal.

    bool displayWatch_;			    ///< Display watch on progress or not.
	
    virtual void on_show();

    void OnBtnPressEvent(GdkEventButton*);
	
    void OnRowActivated(const Gtk::TreeModel::Path&, Gtk::TreeViewColumn*);
    void OnRowExpanded(const Gtk::TreeModel::iterator&, const Gtk::TreeModel::Path&);

    void ExpandCurrentDir();

    void OnMenuRefresh();
    bool GetUnregisteredSubDirectories(std::vector<std::string>& dest);

    void InsertName(Gtk::TreeModel::Row& child, const std::string& name);

    /// @brief Add root directory to tree model.
    void AddRootDirectory()
    {
        Gtk::TreeModel::Row parent( *(treeModel_->append()) );
        parent[column_.colName_] = "/";
        parent[column_.colPath_] = "/";

        treeModel_->append(parent.children());
    }

    bool ThirdButtonPressEvent(const GdkEventButton* const ev)
    {
        if((ev->type == GDK_BUTTON_PRESS) && (ev->button == 3))
            return true;

        return false;
    }

    void ChangeWatchCursor()
    {
        if(displayWatch_) {
            realize();
            get_window()->set_cursor(Gdk::Cursor(Gdk::WATCH));
            while(Glib::MainContext::get_default()->iteration(false));
        }
    }

    void ChangeMainCursor()
    {
        if(displayWatch_)
            get_window()->set_cursor();
    }

    void GetFullPathname(const std::string& pathname)
    {
        fullPath_ = currentPath_;

        if(fullPath_ != "/")
            fullPath_ += '/';

        fullPath_ += pathname;
    }

    void FillRow(Gtk::TreeModel::Row& child, const std::string& name)
    {
        try {

            child[column_.colName_] = Glib::locale_to_utf8(name);

        } catch(Glib::ConvertError& err) {

            child[column_.colName_] = name;

        }
        child[column_.colPath_] = fullPath_;

        if(!access(fullPath_.c_str(), X_OK | R_OK))
            treeModel_->append(child.children());
    }

};

#endif
