#ifndef GUARD_effect_h
#define GUARD_effect_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file effect.h
/// @brief Effector functions header file.

#include <cmath>
#include "preference.h"

namespace Effect {

    class Base {
	public:	
	    explicit Base(): xn1(0.0), xn2(0.0), yn1(0.0), yn2(0.0) {}
	    ~Base() {}

	    void Reset()
	    {
		xn1 = xn2 = yn1 = yn2 = 0.0;
	    }
	    
	    double xn1, xn2, yn1, yn2;
    };

    /// @brief Class for stereo signal.
    class Stereo {
	public:
	    explicit Stereo(const double& l, const double& r): left(l), right(r) {}
	    
	    explicit Stereo(): left(0.0), right(0.0) {}
	    ~Stereo() {}
	    
	    double left, right;
    };

    /// @brief Class for biquad coeffecient. 
    class BiquadCoeff {
	public:
	    BiquadCoeff(): a0_(0.0), a1_(0.0), a2_(0.0), b1_(0.0), b2_(0.0), out_(0.0),
		sampleRate_(0), freq_(0) {}
			
	    virtual ~BiquadCoeff() {}

	    /// @brief Calculate coeffecients.
	    /// @param sampleRate Sample rate.
	    /// @param freq Frequency.
	    /// @param gain Gain.
	    /// @param Q Shape.
	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q) = 0;

	    /// @brief Apply to sample.
	    void Apply(double& orig, Base& base) const
	    {
		InternalProcess(orig, base);
		
		orig = out_;    
	    }

	    /// @brief Return applied sample.
	    const double& Return(const double& orig, Base& base) const
	    {
		InternalProcess(orig, base);

		return out_;
	    }

	    /// @brief Reset member variable for static declared variable.
	    void Reset()
	    {
		a0_ = a1_ = a2_ = b1_ = b2_ = out_ = 0.0;
		sampleRate_ = freq_ = 0;
	    }

	    /// @brief Return current calculated sample rate.
	    const int& SampleRate() const { return sampleRate_; }
	    /// @brief Return current calculated frequency.
	    const int& Freq() const { return freq_; }

	protected:    
	    double a0_, a1_, a2_, b1_, b2_;
	    mutable double out_;
	    int sampleRate_;
	    int freq_;

	    void InternalProcess(const double& orig, Base& base) const
	    {
		out_ = a0_ * orig + a1_ * base.xn1 + a2_ * base.xn2
		    - b1_ * base.yn1 - b2_ * base.yn2;
		
		base.xn2 = base.xn1;
		base.xn1 = orig;
		base.yn2 = base.yn1;
		base.yn1 = out_;
	    }
    };

    
    /// @brief Class for biquad coeffecient used in low shelf EQ.
    class LowShelf : public BiquadCoeff {
	public:
	    LowShelf() {}
	    virtual ~LowShelf() {}

	    /// @brief Calculate coeffecients.
	    /// @param sampleRate Sample rate.
	    /// @param freq Frequency.
	    /// @param gain Gain.
	    /// @param Q Shape.
	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;

		const double omega = 2.0 * M_PI * freq_ / sampleRate_;	
		
		const double sn = sin(omega);
		const double cs = cos(omega);

		const double a = pow(10.0, gain * 0.025);
		
		const double beta = sqrt((a * a + 1.0) / Q - pow((a - 1.0), 2.0));

		const double norm = 1.0 / (a + 1.0 + (a - 1.0) * cs + beta * sn);

		a0_ = (a + 1.0 - (a - 1.0) * cs + beta * sn) * a * norm;
		
		a1_ = (a - 1.0 - (a + 1.0) * cs) * 2.0 * a * norm;

		a2_ = (a + 1.0 - (a - 1.0) * cs - beta * sn) * a * norm;

		b1_ = (a - 1.0 + (a + 1.0) * cs) * -2.0 * norm;

		b2_ = (a + 1.0 + (a - 1.0) * cs - beta * sn) * norm;
		
	    }

    };	    

    /// @brief Class for biquad coeffecient used in high shelf EQ.
    class HighShelf : public BiquadCoeff
    {
	public:
	    HighShelf() {}
	    virtual ~HighShelf() {}

	    /// @brief Calculate coeffecients.
	    /// @param sampleRate Sample rate.
	    /// @param freq Frequency.
	    /// @param gain Gain.
	    /// @param Q Shape.
	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;
		
		const double omega = 2.0 * M_PI * freq_ / sampleRate_;	

		const double sn = sin(omega);
		const double cs = cos(omega);

		const double a = pow(10.0, gain * 0.025);

		const double beta = sqrt((a * a + 1.0) / Q - pow((a - 1.0), 2.0));
		
		const double norm = 1.0 / (a + 1.0 - (a - 1.0) * cs + beta * sn);

		b1_ = (a - 1.0 - (a + 1.0) * cs) * 2.0 * norm;

		b2_ = (a + 1.0 - (a - 1.0) * cs - beta * sn) * norm;
		
		a0_ = (a + 1.0 + (a - 1.0) * cs + beta * sn) * a * norm;

		a1_ = (a - 1.0 + (a + 1.0) * cs) * -2.0 * a * norm;

		a2_ = (a + 1.0 + (a - 1.0) * cs - beta * sn) * a * norm;
		
	    }
    };

    /// @brief Class for biquad coeffecient used in peak EQ.
    class Peak : public BiquadCoeff
    {
	public:
	    Peak() {}
	    virtual ~Peak() {}

	    /// @brief Calculate coeffecients.
	    /// @param sampleRate Sample rate.
	    /// @param freq Frequency.
	    /// @param gain Gain.
	    /// @param Q Shape. Maybe bandwidth.
	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;

		const double omega = 2.0 * M_PI * freq_ / sampleRate_;
		
		const double sn = sin(omega);
		const double cs = cos(omega);

		const double a = pow(10.0, gain * 0.025);

		const double alpha = sn / (2.0 * Q);
		
		const double norm = 1.0 / (alpha / a + 1.0);
		
		b1_ = -2.0 * cs * norm;
		
		b2_ = (1.0 - (alpha / a)) * norm;
		
		a0_ = (1.0 + alpha * a) * norm;
		
		a1_ = -2.0 * cs * norm;
		
		a2_ = (1.0 - alpha * a) * norm;

	    }
    };

    /// @brief Class for biquad coeffecient used in low pass filter. 
    class LowPass : public BiquadCoeff
    {
	public:
	    LowPass() {}
	    virtual ~LowPass() {}

	    /// @brief Calculate coeffecients.
	    /// @param sampleRate Sample rate.
	    /// @param freq Frequency.
	    /// @param gain Gain.
	    /// @param Q Shape.
	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;

		const double omega =  1.0 / tan(M_PI * freq_ / sampleRate_);

		a0_ = 1.0 / (1.0 + (gain + omega) * omega);
		
		a1_ = 2.0 * a0_;
		
		a2_ = a0_;

		b1_ = 2.0 * (1.0 - omega * omega) * a0_;

		b2_ = (1.0 - (gain - omega) * omega) * a0_;
		
	    }
    };

    /// @brief Class for biquad coeffecient used in high pass filter. 
    class HighPass : public BiquadCoeff
    {
	public:
	    HighPass() {}
	    virtual ~HighPass() {}

	    /// @brief Calculate coeffecients.
	    /// @param sampleRate Sample rate.
	    /// @param freq Frequency.
	    /// @param gain Gain.
	    /// @param Q Shape.
	    virtual void Calc(const int& sampleRate, const int& freq, const float& gain, const double& Q)
	    {
		sampleRate_ = sampleRate;
		freq_ = freq;
		
		const double omega = tan(M_PI * freq_ / sampleRate_);

		a0_ = 1.0 / (1.0 + (gain + omega) * omega);

		a1_ = -2.0 * a0_;

		a2_ = a0_;

		b1_ = 2.0 * (omega * omega - 1.0) * a0_;

		b2_ = (1.0 - (gain - omega) * omega) * a0_;

	    }
    };

    void Reset();

    void Process(const Preference& preference, const int& sampleRate, Stereo& orig);
    
    void Clip(double& sample);
    void Clip(double& sampleL, double& sampleR);
    
    void Amplify(double& sampleL, double& sampleR, const float& gain);
    void AmplifyClip(double& sampleL, double& sampleR, const float& gain);
    void AmplifyDecibel(double& sampleL, double& sampleR, const float& gain);

    void BassBoost(const int& sampleRate, double& sampleL, double& sampleR);
    void BassBoost2(const int& sampleRate, double& sampleL, double& sampleR);
    void TrebleBoost(const int& sampleRate, double& sampleL, double& sampleR);
    
    void EQ(BiquadCoeff& peak, const int& sampleRate, double& sampleL, double& sampleR, const int& freq, const float& gain, Base& prevL, Base& prevR, const float& Q = 1.0);
    void GraphicEQ(const int& sampleRate, double& sampleL, double& sampleR, const float* const parameter, const bool& needCalc = false);
    
    Stereo LPF(BiquadCoeff& p, const int& sampleRate, double& sampleL, double& sampleR, const int& freq, const double& gain, Base& prevL, Base& prevR);
    Stereo HPF(BiquadCoeff& p, const int& sampleRate, double& sampleL, double& sampleR, const int& freq, const double& gain, Base& prevL, Base& prevR);
    
    void StereoEnhance(double& left, double& right, const float& wide);
    void Rhythme(double& sampleL, double& sampleR, const double& mount);
    void Room(const int& sampleRate, double& sampleL, double& sampleR, const double& ab);
    void Room2(const int& sampleRate, double& sampleL, double& sampleR, const double& mount, const bool& needCalc = false);
    void Harmony(const int& sampleRate, double& sampleL, double& sampleR, const double& ab);
    void Distortion(double& sampleL, double& sampleR, const double& mount);

}

#endif

