/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
#ifndef GUARD_mainmanager_h
#define GUARD_mainmanager_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file mainmanager.h
/// @brief Class for main program control mediator header file.

#include <string>
#include <vector>
#include <map>

#include <glibmm.h>
#include <gtkmm/main.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "sockethandler.h"
#include "trayicon.h"
#include "audio.h"
#include "preference.h"
#include "mixer.h"
#include "mainwin.h"

class PrefWnd;

/// @brief Class for main program control mediator. 
///
/// Almost class is connected to MainManager, not to exact class.
class MainManager
{
public:
    const unsigned int audioFormatLittleEndian_;		///< Little endian identification variable in libao.
    const unsigned int audioFormatBigEndian_;		///< Big endian identification variable in libao.
    const unsigned int audioFormatNativeEndian_;		///< Machine native endian identification variable in libao.

    /// @brief Create MainManager object. (singleton pattern)
    /// @param arguments Command line arguments.
    /// @param socketHandler UnixSocketHandler object to accept a connection.
    /// @return Pointer of created MainManager object.
    static MainManager* Create(const std::vector<std::string>& arguments, UnixSocketHandler& socketHandler)
    {
        static MainManager mainManager_(arguments, socketHandler);

        return &mainManager_;
    }

    /// @brief Run main loop;
    void Run()
    {
        Gtk::Main::run();
    }


    /// @brief Get readable Preference.
    const Preference& GetPreference() const
    {
        return preference_;
    }

    /// @brief Get writable Preference.
    Preference& GetWritablePreference()
    {
        return preference_;
    }

    /// @brief Get assume charset used when machine's charset is UTF-8.
    /// @sa Preference::GeneralAssumeCharset
    const std::string& GetAssumeCharset() const
    {
        return preference_.GeneralAssumeCharset();
    }

    /// @brief Get Mixer device from Preference.
    /// @sa Preference::MixerDev
    const std::string& MixerDev() const
    {
        return preference_.MixerDev();
    }

#ifdef ENABLE_HTTP
    bool ConvertUrlToCurrentEncoding() const
    {
        return !isCharSetUTF8_ && preference_.ConvertUrlToCurrentEncoding();
    }
#endif

    /// @brief Get audio driver name from Preference.
    /// @sa Preference::AudioDriver
    const std::string& GetAudioDriver() const
    {
        return preference_.AudioDriver();
    }

    /// @brief Set changed Audio driver.
    ///
    /// If player is playing, just stop play.
    /// Otherwise set Audio driver from Preference.
    void AudioSetChangedDriver()
    {
        if(mainWnd_.InPlay())
            mainWnd_.StopPlayForSetDriver();
        else
            audio_->SetDriverFromPreference();
    }

    /// @brief Set Audio driver from Preference.
    /// @sa Audio::SetDriverFromPreference
    void AudioSetDriverFromPreference()
    {
        audio_->SetDriverFromPreference();
    }

    /// @brief Open Audio device.
    /// @param bits Bits per sample.
    /// @param channels Number of channels.
    /// @param sampleRate Sample rate per sample.
    /// @param endian Byte order.
    /// @return Success of open.
    /// @sa Audio::OpenDevice
    bool AudioOpenDevice(const unsigned int& bits, const unsigned int& channels, const unsigned int& sampleRate, const unsigned int& endian)
    {
        if(audio_->OpenDevice(bits, channels, sampleRate, endian))
            return true;
        else {

            signalAudioDeviceOpenFailed_();

            return false;
        }
    }

    /// @brief Device is opened or not.
    /// @sa Audio::CheckOpenDevice
    bool AudioCheckOpenDevice()
    {
        return audio_->CheckOpenDevice();
    }

    /// @brief Close audio device.
    /// @param bufferFlush Flush audio buffer when close device. Default true.
    /// @sa Audio::Close
    void AudioClose(bool bufferFlush = true)
    {
        audio_->Close(bufferFlush);
    }

    /// @brief Test audio driver for driver name.
    /// @param driverName Driver name.
    /// @return Success of test.
    /// @sa Audio::TestDriver(const char* const driverName)
    bool AudioTestDriver(const char* const driverName)
    {
        return audio_->TestDriver(driverName);
    }

    /// @brief Test audio driver for driver name with option.
    /// @param driverName Driver name.
    /// @param options Options.
    /// @return Success of test.
    /// @sa Audio::TestDriver(const char* const driverName, const std::map<std::string, std::string>& options)
    bool AudioTestDriver(const char* const driverName, const std::map<std::string, std::string>& options)
    {
        return audio_->TestDriver(driverName, options);
    }

    /// @brief Play samples.
    /// @param samples Samples to play.
    /// @param bytes Number of sample.
    /// @sa Audio::Play(const std::vector<double>& samples, const unsigned int& bytes)
    void AudioPlay(const std::vector<double>& samples, const unsigned int& bytes)
    {
        audio_->Play(samples, bytes);
    }

    /// @brief Play samples.
    /// @param samples Samples to play.
    /// @param bytes Number of sample.
    /// @sa Audio::Play(float** const samples, const unsigned int& bytes)
    void AudioPlay(float** const samples, const unsigned int& bytes)
    {
        audio_->Play(samples, bytes);
    }

    /// @brief Play samples.
    /// @param samples Samples to play.
    /// @param bytes Number of sample.
    /// @sa Audio::Play(double* const samples, const unsigned int& bytes)
    void AudioPlay(double* const samples, const unsigned int& bytes)
    {
        audio_->Play(samples, bytes);
    }

    /// @brief Toggle pause state.
    /// @sa Audio::PauseToggle
    void AudioPauseToggle()
    {
        audio_->PauseToggle(false);
    }

    /// @brief Unpause audio buffer play.
    ///
    /// Unpause audio buffer play and skip play buffered sample.
    /// @sa Audio::PauseToggle
    void AudioUnpause()
    {
        if(audio_->Paused())
            audio_->PauseToggle(true);
    }

    void SetCurrentTime(const long int& seconds)
    {
        audio_->SetCurrentTime(seconds);
    }

    const long int& GetCurrentTime()
    {
        return audio_->GetCurrentTime();
    }

    /// @brief Refresh Mixer widget.
    /// @sa MainWnd::RefreshMixer
    void MixerRefresh()
    {
        mainWnd_.RefreshMixer();
    }

    /// @brief Get current PCM volumn.
    /// @sa Mixer::GetVolumn
    const double& MixerGetVolumn()
    {
        return mixer_->GetVolumn();
    }

    /// @brief Set PCM volumn.
    /// @param volumn Value to set. 0 ~ 100.
    /// @sa Mixer::SetVolumn
    void MixerSetVolumn(const double& volumn)
    {
        mixer_->SetVolumn(volumn);
    }

    /// @brief Mixer Volumn increase.
    /// @sa MainWnd::VolumnIncrease 
    void MixerVolumnIncrease()
    {
        mainWnd_.VolumnIncrease();
    }

    /// @brief Mixer Volumn decrease.
    /// @sa MainWnd::VolumnDecrease
    void MixerVolumnDecrease()
    {
        mainWnd_.VolumnDecrease();
    }


    /// @brief Play previous position file at playlist.
    /// @sa MainWnd::OnBtnPrev
    void PlayCtrlPrev()
    {
        mainWnd_.OnBtnPrev();
    }

    /// @brief Play file.
    /// @sa MainWnd::OnBtnPlay
    void PlayCtrlPlay()
    {
        mainWnd_.OnBtnPlay();
    }

    /// @brief Stop play.
    /// @sa MainWnd::OnBtnStop
    void PlayCtrlStop()
    {
        mainWnd_.OnBtnStop();
    }

    /// @brief Play next position file at playlist.
    /// @sa MainWnd::OnBtnNext
    void PlayCtrlNext()
    {
        mainWnd_.OnBtnNext();
    }

    /// @brief Callback function used when play is finished.
    /// @sa MainWnd::FinishPlay
    void PlayFinish()
    {
        mainWnd_.FinishPlay();
    }

    /// @brief Callback function used when player need to update timer.
    /// @sa MainWnd::TimeUpdate
    void PlayTimeUpdate()
    {
        mainWnd_.TimeUpdate();
    }


    /// @brief Show current file information.
    /// @sa MainWnd::CurrentFileInfo
    void FileInfo()
    {
        mainWnd_.CurrentFileInfo();
    }


    /// @brief Set tooltips for trayicon.
    /// @sa TrayIcon::SetTooltips
    void TrayIconSetTooltips(const Glib::ustring& string)
    {
        trayIcon_.SetTooltips(string);
    }

    /// @brief Callback function when trayicon is pressed by 1 button.
    void TrayIconFirstBtnClickAction()
    {
        CallbackCIter i(trayIconFirstBtnCallbacks_.find(preference_.GeneralTrayIconAction()));

        if(i != trayIconFirstBtnCallbacks_.end())
            (this->*(i->second))();
    }

    void MainWndOpen();
    void MainWndChangeCompactBgPixmap()
    {
        mainWnd_.ChangeCompactBgPixmap();
    }

    /// @brief Add received file list to playlist.
    void SocketReceived()
    {
        std::vector<std::string> received;

        socketHandler_.ReceivedData(received);

        mainWnd_.PushNewListFromSocket(received);

    }

    /// @brief Set MainWnd statusbar text and update it.
    /// @sa MainWnd::SetStatusbar
    void SetStatusbar(const Glib::ustring& text)
    {
        mainWnd_.SetStatusbar(text);
    }

    /// @brief Set MainWnd statusbar text.
    /// @sa MainWnd::SetStatusbarText
    void SetStatusbarText(const Glib::ustring& text)
    {
        mainWnd_.SetStatusbarText(text);
    }

    /// @brief Update MainWnd statusbar to stored text.
    /// @sa MainWnd::UpdateStatusbar
    void UpdateStatusbar()
    {
        mainWnd_.UpdateStatusbar();
    }

    /// @brief Update MainWnd statusbar to stored text and clear at timeout.
    /// @sa MainWnd::UpdateStatusbarTimeoutClear
    void UpdateStatusbarTimeoutClear()
    {
        mainWnd_.UpdateStatusbarTimeoutClear();
    }

    /// @brief Clear MainWnd statusbar.
    /// @sa MainWnd::ClearStatusbar
    void ClearStatusbar()
    {
        mainWnd_.ClearStatusbar();
    }

    /// @brief Clear MainWnd statusbar at timeout.
    /// @sa MainWnd::TimeoutClearStatusbar
    void TimeoutClearStatusbar()
    {
        mainWnd_.TimeoutClearStatusbar();
    }

    void About();

    void Quit();
	
    const std::string& GetProgramDir() const
    {
        return programDir_;
    }

    std::string GetProgramDirFilePath(const std::string& file) const
    {
        return programDir_ + file;
    }

    bool CheckProgramDir() const
    {
        if(!Glib::file_test(programDir_, Glib::FILE_TEST_EXISTS)) {
            if(!mkdir(programDir_.c_str(), S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH))
                return false;

        } else if(!Glib::file_test(programDir_, Glib::FILE_TEST_IS_DIR))
            return false;
	    
        return true;
    }

    /// @brief Convert string to UTF8 interface function.
    ///
    /// Try to convert 'orig' from machine's charset to UTF8. 
    /// If convert is failed, retry to convert 'fail' to UTF8.
    ///
    /// If machine's charset is UTF8 and 'orig' is not UTF8, try to convert
    /// 'orig' from 'assume' charset to UTF8.
    ///
    /// @param orig Original string to convert.
    /// @param fail Retry string when convert is failed.
    /// @return Converted string.
    Glib::ustring StringToUTF8(const Glib::ustring& orig, const std::string& fail) const;

    /// @brief Convert string to UTF8 interface function without fail string.
    /// @param orig Original string to convert.
    /// @return Converted string.
    /// @sa MainManager::StringToUTF8(const Glib::ustring& orig, const std::string& fail)
    Glib::ustring StringToUTF8(const Glib::ustring& orig) const;

private:
    MainManager(const std::vector<std::string>& arguments, UnixSocketHandler& unixSocketHandler);
    ~MainManager();
    MainManager(const MainManager& mainManager);
    MainManager& operator=(const MainManager& mainManager);

    const std::string programDir_;
    std::string currentCharSet_;

    const bool isCharSetUTF8_;

    UnixSocketHandler& socketHandler_;
    TrayIcon trayIcon_;
    Preference preference_;
    std::auto_ptr<Audio> audio_;
    Mixer* mixer_;
    MainWnd mainWnd_;

    Glib::Dispatcher signalAudioDeviceOpenFailed_;

    typedef std::map<std::string, void (MainManager::*)()> Callback;
    typedef Callback::const_iterator CallbackCIter;

    Callback trayIconFirstBtnCallbacks_;

    /// @brief Convert string to UTF8 without fail string.
    /// @param orig Original string to convert.
    /// @param charset From charset.
    /// @return Converted string.
    Glib::ustring StringToUTF8WithoutFail(const Glib::ustring& orig, const std::string& charset) const
    {
        try {

            return Glib::convert_with_fallback(orig, "UTF-8", charset);

        } catch(Glib::ConvertError& err) {

            return Glib::ustring();

        }
    }

    /// @brief Convert string to UTF8 with fail string.
    /// @param orig Original string to convert.
    /// @param fail Retry string when convert is failed.
    /// @param charset From charset.
    /// @return Converted string.
    /// @sa Glib::ustring StringToUTF8
    Glib::ustring StringToUTF8WithFail(const Glib::ustring& orig, const std::string& fail, const std::string& charset) const
    {
        try {

            return Glib::convert_with_fallback(orig, "UTF-8", charset);

        } catch(Glib::ConvertError& err) {

            return StringToUTF8WithoutFail(fail, charset);

        }
    }
};

#endif
