/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
#ifndef GUARD_mainwin_h
#define GUARD_mainwin_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file mainwin.h
/// @brief Class for main window header file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <string>

#include <glibmm.h>
#include <gdkmm.h>
#include <gtkmm/window.h>
#include <gtkmm/box.h>
#include <gtkmm/label.h>
#include <gtkmm/menu.h>
#include <gtkmm/menubar.h>
#include <gtkmm/checkmenuitem.h>
#include <gtkmm/radiobuttongroup.h>
#include <gtkmm/radiomenuitem.h>
#include <gtkmm/alignment.h>
#include <gtkmm/scale.h>
#include <gtkmm/button.h>
#include <gtkmm/toolbar.h>
#include <gtkmm/toolbutton.h>
#include <gtkmm/scrolledwindow.h>
#include <gtkmm/eventbox.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/statusbar.h>

#include "imagebutton.h"
#include "player.h"

class MainManager;
class Player;
class PlayerFactory;
class Playlist;

/// @brief Class for main window.
class MainWnd : public Gtk::Window
{
    friend class Playlist;
    
public:
    MainWnd(MainManager& mainManager, const std::vector<std::string>& arguments);
    ~MainWnd();

    void FinishPlay();
    void TimeUpdate();

    /// @brief If filename is playable, return true.
    bool IsPlayable(const std::string& filename) const
    {
        return playerFactory_->Exist(filename);
    }

    /// @brief Store last browsed directory in DirPlayWnd.
    void LogBrowsedDirectory(const std::string& dir)
    {
        lastBrowsedDirectory_ = dir;
    }

    /// @brief Get last browsed directory.
    const std::string& GetLastBrowsedDirectory() const 
    {
        return lastBrowsedDirectory_;
    }

    void OnBtnPlay();
    void OnBtnNext();
    void OnBtnPrev();
    void OnBtnStop();

    void OnMenuPreference();

    void OnMenuQuit();

    /// @brief Thread is running or not.
    const bool& InPlay() const
    {
        return inThread_;
    }
	
    /// @brief Stop play for set changed driver.
    ///
    /// Stop play and call MainManager::AudioSetDriverFromPreference to 
    /// set changed driver in MainWnd::FinishPlay. 
    void StopPlayForSetDriver()
    {
        if(!breakPlayForSetDriver_) {
            breakPlayForSetDriver_ = true;

            OnBtnStop();
        }
    }

    void CurrentFileInfo();

    void SaveState();

    /// @brief Set to skip time display update.
    void SetSkipTimeUpdate(const bool& active) { skipTimeUpdate_ = active; }

    void RefreshMixer();
    
    /// @brief Change mixer widget to increased value and set volumn to that value.
    void VolumnIncrease()
    {
        SetVolumn(IncreasedVolumnValue(volumnScale_.get_value()));
    }

    /// @brief Change mixer widget to decreased value and set volumn to that value.
    void VolumnDecrease()
    {
        SetVolumn(DecreasedVolumnValue(volumnScale_.get_value()));
    }

    void PushNewListFromSocket(const std::vector<std::string>& list);

    /// @brief Change background pixmap used when compact mode.
    void ChangeCompactBgPixmap()
    {
        if(compactPlayMenu_.get_active())
            SetBackgroundPixmapInCompactMode();
    }

    void PlayQueue()
    {
        if(playQueue_) {
            playQueue_ = false;
            OnBtnPlay();
        }
    }

    /// @brief Set statusbar text and update it.
    void SetStatusbar(const Glib::ustring& text)
    {
        DisconnectTimeoutClearStatusbar();

        statusbar_.pop();
        statusbar_.push(text);
    }

    /// @brief Set statusbar text and update it.

    /// @brief Set statusbar text.
    ///
    /// For inter thread dispatcher.
    void SetStatusbarText(const Glib::ustring& text)
    {
        statusbarText_ = text;
    }

    /// @brief Update statusbar to stored text.
    ///
    /// For inter thread dispatcher.
    /// @sa SetStatusbarText
    void UpdateStatusbar()
    {
        DisconnectTimeoutClearStatusbar();

        statusbar_.pop();
        statusbar_.push(statusbarText_);
    }

    /// @brief Update statusbar to stored text and clear at timeout.
    void UpdateStatusbarTimeoutClear()
    {
        UpdateStatusbar();
        TimeoutClearStatusbar();
    }

    /// @brief Clear statusbar.
    void ClearStatusbar()
    {
        statusbar_.pop();
    }

    /// @brief Clear statusbar at timeout.
    void TimeoutClearStatusbar()
    {
        // need this? anyway double check it.
        DisconnectTimeoutClearStatusbar();

        clearStatusbarTimeout_ = new sigc::connection(Glib::signal_timeout().connect(sigc::mem_fun(*this, &MainWnd::TimeoutClearStatusbarSlot), 3000));
    }

    void OnAudioDeviceOpenFailed();

private:
    Gtk::FileChooserDialog fileChooserDlg_;
    Gtk::VBox mainBox_, playerBox_, compactPlayCtlBox_;
    Gtk::HBox playCtlBox_, compactPlayCtlHBox_, timeDisplayBox_, playerBoxWithVolumn_, viewBox_;
    Gtk::MenuBar menuBar_;
    Gtk::Menu fileMenu_, editMenu_, viewMenu_, helpMenu_;
    Gtk::CheckMenuItem randomPlayMenu_, repeatPlayMenu_, compactPlayMenu_, statusbarDisplayMenu_;
    Gtk::RadioButtonGroup viewMenuGroup_;
    Gtk::RadioMenuItem mainlistPlayMenu_, dirPlayMenu_;
    Gtk::Label displayLabel_, timeLabel_, totalTimeLabel_;
    Gtk::Alignment titleDisplayAlignment_, seekScaleAlignment_, volumnAlignment_;
    Gtk::EventBox titleDisplayEventBox_;
    Gtk::HScale seekScale_;
    Gtk::VScale volumnScale_;
    ImageButton playImgBtn_, pauseImgBtn_, stopImgBtn_, prevImgBtn_, nextImgBtn_;
    ImageButton compactPlayImgBtn_, compactPauseImgBtn_, compactStopImgBtn_, compactPrevImgBtn_, compactNextImgBtn_;

    Gtk::Statusbar statusbar_;
    Glib::ustring statusbarText_;

    std::vector<std::string> selectedFilesVector_;

    std::string selectedFile_;
    std::string lastBrowsedDirectory_;

    const int defaultWidth_, defaultHeight_;
    const int defaultCompactWidth_, defaultCompactHeight_;
    int prevWidth_, prevHeight_;

    bool inThread_;
    bool inPlay_;
    bool breakPlay_;
    bool breakPlayForSetDriver_;
    bool randomPlay_;
    bool repeatPlay_;

    bool skipTimeUpdate_;

    bool playQueue_;

    enum ListState {
        LISTSTATE_EMPTY = 3,
        LISTSTATE_LAST = 2,
        LISTSTATE_FIRST = 1,
        LISTSTATE_NORMAL = 0
    };

    ListState listState_;

    Glib::RefPtr<Gtk::RcStyle> mainWndDefaultStyle_;
    Glib::RefPtr<Gtk::RcStyle> titleDisplayEventBoxDefaultStyle_;

    std::auto_ptr<Player> player_;
    std::auto_ptr<Gtk::HBox> currentViewBox_;
    std::auto_ptr<PlayerFactory> playerFactory_;

    Playlist* playlist_;
    Glib::Thread* execThread_;
    Gtk::ScrolledWindow* scrolledWindow_;

    sigc::connection* clearStatusbarTimeout_;

    MainManager& mainManager_;

    virtual bool on_delete_event(GdkEventAny*);
    virtual bool on_map_event(GdkEventAny* event);
    virtual bool on_unmap_event(GdkEventAny* event);

    void OnMenuFileAdd();
    void OnMenuFileDirAdd();
#ifdef ENABLE_HTTP
    void OnMenuFileUrlAdd();
#endif
    void OnMenuRepeat();
    void OnMenuRandom();
    void OnMenuViewMainList();
    void OnMenuViewDir();
    void OnMenuViewStatusbar();
    void OnMenuCompactView();
    void OnMenuViewHide();

    void OnScaleSeek(const double&);

    bool OnButtonPressInDisplayLabel(GdkEventButton* ev);

    void SetBackgroundPixmapInCompactMode();
    void UnsetBackgroundPixmapInCompactMode();

    void LoadMainList(const bool& loadListFile);

    void ClearDisplayTitle();
    void DisplayTitle(const Glib::ustring& title);

    /// @brief Display total time part.
    void DisplayTotalTime(const Glib::ustring& time)
    {
        totalTimeLabel_.set_label(" / " + time);
    }

    /// @brief Display time and set sensitive to seek bar or not.
    void DisplayVisualPanel(const bool& flag)
    {
        if(flag) {
            timeLabel_.show();
            totalTimeLabel_.show();
            seekScale_.set_sensitive(true);	
        } else {
            timeLabel_.hide();
            totalTimeLabel_.hide();
            seekScale_.set_sensitive(false);
        }	
    }

    /// @brief Set maximum range of seek bar.
    void SetSeekScaleRange(const unsigned int& max)
    {
        seekScale_.set_range(0.0, max);
    }

    void LoadLastState(const bool& hasArguments);

    const bool& GetRepeatPlay() const
    {
        return repeatPlay_;
    }

    void SetListState(const ListState& listState)
    {
        listState_ = listState;
    }

    const std::string& GetPlayFilename() const
    {
        return selectedFile_;
    }

    void SetPlayFilename(const std::string& filename)
    {
        selectedFile_ = filename;
    }

    const int& GetPrevHeight() const
    {
        return prevHeight_;
    }

    void RegisterPlayerFactory();
    void SetFileChooserDlgProperty();
    void CreateMenuBar();
    void SetMouseCursorToWatch()
    {
        get_window()->set_cursor(Gdk::Cursor(Gdk::WATCH));
        while(Glib::MainContext::get_default()->iteration(false));
    }

    void RestoreMouseCursor()
    {
        get_window()->set_cursor();
    }

    void EmptySelectedFilename()
    {
        OnBtnStop();
        selectedFile_ = "";
        ClearDisplayTitle();
    }

    double IncreasedVolumnValue(const double& orig) const
    {
        if(orig > 95.0)
            return 100.0;
	    
        return orig + 5.0;
    }

    double DecreasedVolumnValue(const double& orig) const
    {
        if(orig < 5.0)
            return 0.0;

        return orig - 5.0;
    }

    void SetVolumn(const double& value);

    Glib::ustring BoldWeightTitle(const Glib::ustring& title, const bool& normalSize);

    const MainManager& GetMainManager() const
    {
        return mainManager_;
    }

    void SetDisplayLabel(const Glib::ustring& string)
    {
        displayLabel_.set_label(string);
        displayLabel_.set_use_markup(true);
        displayLabel_.set_alignment(0.5, 0.5);
    }

    /// @brief Disconnect timeout handler and delete it.
    void DisconnectTimeoutClearStatusbar()
    {
        if(clearStatusbarTimeout_ == NULL)
            return;

        clearStatusbarTimeout_->disconnect();
        delete clearStatusbarTimeout_;
        clearStatusbarTimeout_ = NULL;
    }

    /// @brief Clear statusbar function for timeout handler.
    bool TimeoutClearStatusbarSlot()
    {
        ClearStatusbar();

        delete clearStatusbarTimeout_;
        clearStatusbarTimeout_ = NULL;

        return false;
    }
};

#endif
