/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
#ifndef GUARD_misc_h
#define GUARD_misc_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file misc.h
/// @brief Miscellaneous functions.
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <vector>
#include <map>
#include <set>
#include <string>
#include <cstdio>
#include <cctype>

#include <glibmm.h>

/// @brief Function object to convert char to lower case.
struct ToLower: std::unary_function<char, void>
{
    void operator()(char& p) const
    {
        p = std::tolower(p);
    }
};

/// @brief Function object to convert char to upper case.
struct ToUpper: std::unary_function<char, void>
{
    void operator()(char& p) const
    {
        p = std::toupper(p);
    }
};

/// @brief Function object to test char is white-space character.
struct IsCharSpace: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
        return isspace(p);	
    }
};    

/// @brief Function object to test char is '='.
struct IsCharEqual: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
        if(p == '=')
            return true;
        else
            return false;
    }
};

/// @brief Function object to test char is '/'.
struct IsCharSlash: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
        if(p == '/')
            return true;
        else
            return false;    
    }
};

/// @brief Function object to test char is '.'.
struct IsCharComa: std::unary_function<char, bool>
{
    bool operator()(const char& p) const
    {
        if(p == '.')
            return true;
        else
            return false;    
    }
};

/// @brief Test string is just white-space.
inline bool StringIsSpace(const Glib::ustring& str)
{
    return find_if(str.begin(), str.end(), std::not1(IsCharSpace())) == str.end();
}

/// @brief Test path is directory.
inline bool IsDirectory(const std::string& path)
{
    return Glib::file_test(path, Glib::FILE_TEST_IS_DIR);
}

/// @brief Test path is executable.
///
/// See also Glib::file_test
inline bool IsExecutable(const std::string& path)
{
    return (access(path.c_str(), X_OK) == 0);
}

/// @brief Test path is readable.
inline bool IsReadable(const std::string& path)
{
    return (access(path.c_str(), R_OK) == 0);
}

std::string GetExt(const std::string& filename);

/// @brief Microseconds interval sleep.
///
/// This use select function.
inline int usecsleep(const int& secs, const int& usecs)
{
    struct timeval tv;

    tv.tv_sec = secs;
    tv.tv_usec = usecs;

    return select(0, NULL, NULL, NULL, &tv);
}

inline Glib::ustring ConvertSecondToTimeString(const unsigned long& second)
{
    char timeStr[18] = { 0, };

    if(second < 3600)
        std::snprintf(timeStr, sizeof(timeStr), "%02lu:%02lu", second / 60, second % 60);
    else
        std::snprintf(timeStr, sizeof(timeStr), "%02lu:%02lu:%02lu", second / 3600, second % 3600 / 60, second % 3600 % 60);
    
    return timeStr;
}

inline unsigned int StringHash(const std::string& orig)
{
    const unsigned int size = orig.size();
    unsigned int h = 0;
    for(unsigned int i = 0; i < size; ++i) {

            h = 5 * h + orig[i];
    }

    return h;
}

inline bool HasHttpProtocolString(const std::string& filePath)
{
    if(filePath.size() < 7)
        return false;

    // convert to lower case and strcmp ?

    if((filePath[0] == 'h' || filePath[0] == 'H') &&
       (filePath[1] == 't' || filePath[1] == 'T') &&
       (filePath[2] == 't' || filePath[2] == 'T') &&
       (filePath[3] == 'p' || filePath[3] == 'P') &&
       filePath[4] == ':' && filePath[5] == '/' &&
       filePath[6] == '/')
        return true;
    else
        return false;
}

void GetFilenameFromDirToVector(const std::string& path, std::vector<std::string>& dest, bool recursive = true);

bool ConfigfileToMap(const std::string& filename, std::map<std::string, std::string>& configMap, const bool& convertLowerCase = true, const bool& valueBreakAtSpace= true);

#endif
