#ifndef GUARD_preference_h
#define GUARD_preference_h
/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file preference.h
/// @brief Class for user preference setting header file.

#include <map>
#include <string>

class Preference;
class PrefWnd;

/// @brief Class for user preference setting.
class Preference {
public:
    Preference();
    ~Preference();

    friend void operator<<(Preference& preference, const PrefWnd& prefWnd);

    /// @brief Success load previous config file or not.
    const bool& SuccessLoadConfig() const { return successLoadConfig_; }

#ifdef ENABLE_HTTP
    const bool& ConvertUrlToCurrentEncoding() const { return convertUrlToCurrentEncoding_; }
#endif

    /// @brief Get action used when tray icon is clicked.
    const std::string& GeneralTrayIconAction() const { return generalTrayIconAction_; }
    /// @brief Assume charset when machine's charset is UTF-8.
    const std::string& GeneralAssumeCharset() const { return generalAssumeCharset_; }
    /// @brief Background pixmap when MainWnd is compact mode.
    const std::string& GeneralCompactBgPixmap() const { return generalCompactBgPixmap_; }

    /// @brief Get selected audio driver name.
    const std::string& AudioDriver() const { return audioDriver_; }
    /// @brief Get default audio driver name.
    void SetDefaultAudioDriver() { audioDriver_ = "default"; }
	
    /// @brief Get selected option of alsa v0.5.x card section. 
    const std::string& AudioAlsa5Card() const { return audioAlsa5Card_; }
    /// @brief Get selected option of alsa v0.5.x device section. 
    const std::string& AudioAlsa5Dev() const { return audioAlsa5Dev_; }

    /// @brief Get selected option of alsa v0.9.x device section. 
    const std::string& AudioAlsa9Dev() const { return audioAlsa9Dev_; }

    /// @brief Get selected option of OSS device section. 
    const std::string& AudioOssDsp() const { return audioOssDsp_; }
    /// @brief Get selected option of sun device section. 
    const std::string& AudioSunDev() const { return audioSunDev_; }
    /// @brief Get selected option of aixs device section. 
    const std::string& AudioAixsDev() const { return audioAixsDev_; }

    /// @brief Get mixer device path. 
    const std::string& MixerDev() const { return mixerDev_; }

    /// @brief Get amount of stereo enhance effector option. 
    const double& EfxStereoEnhanceAmount() const { return efxStereoEnhanceAmount_; }
    /// @brief Get amount of room effector option. 
    const double& EfxRoomAmount() const { return efxRoomAmount_; }
    /// @brief Get amount of room2 effector option. 
    const double& EfxRoom2Amount() const { return efxRoom2Amount_; }
    /// @brief Get amount of harmony effector option. 
    const double& EfxHarmonyAmount() const { return efxHarmonyAmount_; }
    /// @brief Get amount of rhythme effector option. 
    const double& EfxRhythmeAmount() const { return efxRhythmeAmount_; }
    /// @brief Get amount of distortion effector option. 
    const double& EfxDistortionAmount() const { return efxDistortionAmount_; }

    /// @brief Get values of graphic EQ.
    /// @return Array of float, size 11.
    const float* EfxGraphicEQGain() const { return efxGraphicEQGain_; }

    /// @brief Use effector of not.
    const bool& Effect() const { return effect_; }    
    /// @brief Need to calculate some effector coefficient.
    const bool& EfxChanged() const { return efxChanged_; }
    /// @brief Set efxChanged to true.
    void EfxChangedSet() const { efxChanged_ = true; }
    /// @brief Set efxChanged to false.
    void EfxChangedUnset() const { efxChanged_ = false; }
    /// @brief Use bass boost effector or not.
    const bool& EfxBassBoost() const { return efxBassBoost_; }
    /// @brief Use bass boost2 effector or not.
    const bool& EfxBassBoost2() const { return efxBassBoost2_; }
    /// @brief Use stereo enhance effector or not.
    const bool& EfxStereoEnhance() const { return efxStereoEnhance_; }
    /// @brief Use treble boost effector or not.
    const bool& EfxTrebleBoost() const { return efxTrebleBoost_; }
    /// @brief Use room effector or not.
    const bool& EfxRoom() const { return efxRoom_; }
    /// @brief Use room2 effector or not.
    const bool& EfxRoom2() const { return efxRoom2_; }
    /// @brief Use harmony effector or not.
    const bool& EfxHarmony() const { return efxHarmony_; }
    /// @brief Use rhythme effector or not.
    const bool& EfxRhythme() const { return efxRhythme_; }
    /// @brief Use distortion effector or not.
    const bool& EfxDistortion() const { return efxDistortion_; }
    /// @brief Use graphic EQ or not.
    const bool& EfxGraphicEQ() const { return efxGraphicEQ_; }

private:
    std::string generalTrayIconAction_; 
    std::string generalAssumeCharset_;
    std::string generalCompactBgPixmap_;
    
    std::string audioDriver_;
	
    std::string audioAlsa5Card_;
    std::string audioAlsa5Dev_;

    std::string audioAlsa9Dev_;

    std::string audioOssDsp_;
    std::string audioSunDev_;
    std::string audioAixsDev_;

    std::string mixerDev_;
    
    double efxStereoEnhanceAmount_;
    double efxRoomAmount_;
    double efxRoom2Amount_;
    double efxHarmonyAmount_;
    double efxRhythmeAmount_;
    double efxDistortionAmount_;

    float efxGraphicEQGain_[11];
	
    bool successLoadConfig_;
#ifdef ENABLE_HTTP
    bool convertUrlToCurrentEncoding_;
#endif
    bool effect_;
    mutable bool efxChanged_;
    bool efxBassBoost_;
    bool efxBassBoost2_;
    bool efxStereoEnhance_;
    bool efxTrebleBoost_;     
    bool efxRoom_;
    bool efxRoom2_;
    bool efxHarmony_;
    bool efxRhythme_;
    bool efxDistortion_;
    bool efxGraphicEQ_;

    bool LoadConfig();
    void AssignConfig(bool& dest, const std::string& find, const std::map<std::string, std::string>& configMap);
    void AssignConfig(double& dest, const std::string& find, const std::map<std::string, std::string>& configMap);
    void AssignConfig(float& dset, const std::string& find, const std::map<std::string, std::string>& configMap);
    void AssignConfig(std::string& dset, const std::string& find, const std::map<std::string, std::string>& configMap);
	
};

#endif
