/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
#ifndef GUARD_tag_h
#define GUARD_tag_h
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag.h
/// @brief Abstract class for tag processor and factory class to produce tag 
/// processor object.

#include <string>

#include <glibmm.h>

#include "fixedallocator.h"
#include "misc.h"

class MainManager;

/// @brief Abstract class for tag processor.
class Tag 
{ 
public:
    /// @brief Constructor of Tag.
    ///
    /// Gtk::Scale's minimum value is 0 and maximum value must be bigger
    /// than minimum value. So i'll take 1 to totalTimeSecond initial value
    /// , not 0.
    Tag(): totalTimeSeconds_(1) {}
    virtual ~Tag() {}

    /// @brief Get string of title section.
    const Glib::ustring& GetTitle()
    {
        if(title_.empty())
            ProcessTitle();

        return title_;
    }
	
    /// @brief Get string of album section.
    const Glib::ustring& GetAlbum()
    {
        if(album_.empty())
            ProcessAlbum();

        return album_;
    }

    /// @brief Get string of genre section.
    const Glib::ustring& GetGenre()
    {
        if(genre_.empty())
            ProcessGenre();

        return genre_;
    }
	
    /// @brief Get string of artist section.
    const Glib::ustring& GetArtist()
    {
        if(artist_.empty())
            ProcessArtist();

        return artist_;
    }
	
    /// @brief Get string of comment section.
    const Glib::ustring& GetComment() 
    {
        if(comment_.empty())
            ProcessComment();

        return comment_;	
    }

    /// @brief Get string of year section.
    const Glib::ustring& GetYear() 
    {
        if(year_.empty())
            ProcessYear();
		
        return year_;
    }
	
    /// @brief Get string of track section.
    const Glib::ustring& GetTrack() 
    {
        if(track_.empty())
            ProcessYear();

        return track_;
    }

    /// @brief Get string of total time.
    const Glib::ustring& GetTotalTime()
    {
        if(totalTime_.empty())
            ProcessTotalTime();

        return totalTime_;
    }

    /// @brief Get unsigned integer value of total time in second.
    const unsigned long& GetTotalTimeSeconds()
    {
        if(totalTimeSeconds_ == 1)
            ProcessTotalTime();

        return totalTimeSeconds_;	
    }

protected:
    // To display in window, i'll use Glib::ustring instead std::string
    Glib::ustring title_, artist_, album_, track_, year_;
    Glib::ustring genre_, comment_;
    Glib::ustring totalTime_;
    unsigned long totalTimeSeconds_;

    virtual void ProcessYear() = 0;
    virtual void ProcessTrack() = 0;
    virtual void ProcessTitle() = 0;
    virtual void ProcessAlbum() = 0;
    virtual void ProcessGenre() = 0;
    virtual void ProcessArtist() = 0;
    virtual void ProcessComment() = 0;
    virtual void ProcessTotalTime() = 0;
};

/// @brief Factory class to produe tag processor object.
class TagFactory : public SmallObject
{
public:
    /// @brief Register id to callback map.
    /// @param id Identification string.
    /// @param mainManager MainManager to contain assume charset information and others.
    /// @param creator Callback function to produce object.
    /// @return Success of register.
    bool Register(const std::string& id, MainManager* const mainManager, std::auto_ptr<Tag> (*creator)(const std::string&, const MainManager* const))
    {
        mainManager_ = mainManager;
	    
        return callbacks_.insert( callback::value_type(id, creator) ).second;
    }

    /// @brief Unregister id from callback map.
    /// @param id Identification string.
    /// @return Success of unregister.
    bool Unregister(const std::string& id)
    {
        return callbacks_.erase(id) == 1;
    }

    /// @brief Check factory has filename's type.
    /// @param filename Filename to check.
    /// @return Return true if factory has filename's type.
    bool Exist(const std::string& filename)
    {
        callback::const_iterator i = callbacks_.find(GetExt(filename));
	    
        if(i != callbacks_.end())
            return true;
        else
            return false;	
    }
	
    /// @brief Create tag processor object.
    /// @param filename filename to process.
    /// @return Tag processor object.
    std::auto_ptr<Tag> Create(const std::string& filename)
    {
        callback::const_iterator i = callbacks_.find(GetExt(filename));

        if(i != callbacks_.end())
            return (i->second)(filename, mainManager_);
        else
            return std::auto_ptr<Tag>();	
    }

    /// @brief True if there is nothing registered.
    bool Empty() const
    {
        return callbacks_.empty();
    }

private:
    typedef std::map<std::string, std::auto_ptr<Tag> (*)(const std::string&, const MainManager* const)> callback;
    callback callbacks_;	    ///< Callback map.
    MainManager* mainManager_;
};


#endif
