/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file connector.cc
/// @brief Class for network connector source file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ace/Codecs.h>

#include "../include/connector.h"

#if defined(ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)

template class ACE_Connector<Client, ACE_SOCK_CONNECTOR>;
template class ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_MT_SYNCH>;

#elif defined(ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)

#pragma instantiate ACE_Connector<Client, ACE_SOCK_CONNECTOR>
#pragma instantiate ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_MT_SYNCH>

#endif

/// @brief Namespace for some network utility.
namespace NetUtil {

    const size_t packageInfoLength = ACE_OS::strlen(PACKAGE_NAME) + ACE_OS::strlen(VERSION);

    /// @brief Send HTTP GET request to peer.
    /// @param request Request string, URL at most time.
    /// @param peer Peer to send message.
    /// @param accept Accept media range. All media type at default.
    void SendGetMessage(const ACE_TCHAR* request,
                        ACE_SOCK_Stream& peer,
                        const ACE_TCHAR* accept)
    {
        const size_t messageSize = ACE_OS::strlen(request) + ACE_OS::strlen(accept) + packageInfoLength + 43;
        char* message = new char[messageSize];

        ACE_OS::memset(message, 0, messageSize);
        ACE_OS::snprintf(message,
                         messageSize,
                         "GET %s HTTP/1.0\r\n" \
                         "User-Agent: " PACKAGE_NAME "/" VERSION "\r\n" \
                         "Accept: %s\r\n" \
                         "\r\n",
                         request,
                         accept);

        peer.send(message, messageSize);

        delete [] message;
    }

    /// @brief Send HTTP GET request to peer.
    /// @param request Request string, URL at most time.
    /// @param author Base64 encoded author string.
    /// @param peer Peer to send message.
    /// @param accept Accept media range. All media type at default.
    void SendGetMessage(const ACE_TCHAR* request,
                        const ACE_Byte* author,
                        ACE_SOCK_Stream& peer,
                        const ACE_TCHAR* accept)
    {

        const size_t messageSize = ACE_OS::strlen(request) + ACE_OS::strlen(ACE_reinterpret_cast(const char*, author)) + ACE_OS::strlen(accept) + packageInfoLength + 66;

        char* message = new char[messageSize];

        ACE_OS::memset(message, 0, messageSize);
        ACE_OS::snprintf(message,
                         messageSize,
                         "GET %s HTTP/1.0\r\n" \
                         "User-Agent: " PACKAGE_NAME "/" VERSION "\r\n" \
                         "Accept: %s\r\n" \
                         "Authorization: Basic %s\r\n" \
                         "\r\n",
                         request,
                         accept,
                         author);
        peer.send(message, messageSize);

        delete [] message;
    }

};

const char* Address::httpProtocolString_ = "http://";

Address::Address(const ACE_TCHAR* str)
    : portNumber_(80),
      rawHostName_(NULL),
      request_(NULL),
      fullAddress_(NULL),
      authorizationString_(NULL),
      encodedAuthorizationString_(NULL)
{
    SplitAddressString(str);
}

Address::~Address()
{
    if(encodedAuthorizationString_ != NULL)
        delete [] encodedAuthorizationString_;

    if(authorizationString_ != NULL)
        delete [] authorizationString_;

    if(fullAddress_ != NULL)
        delete [] fullAddress_;

    if(request_ != NULL)
        delete [] request_;

    if(rawHostName_ != NULL)
        delete [] rawHostName_;
}

/// @brief Split original string.
void Address::SplitAddressString(const ACE_TCHAR* str)
{
    // Process address string.
    //
    // Check http protocol string -> Check authorization string ->
    // Check port number string.
    //
    // Possible types. All types are possbile has authorization string.
    //
    // type 1. http://address
    // type 2. http://address:port
    // type 3. http://address/some/where
    // type 4. http://address:port/some/where
    // type 5. address
    // type 6. address:port
    // type 7. address/some/where
    // type 8. address:port/some/where

    if(ACE_OS::strlen(str) > httpProtocolStringLength_ &&
       ACE_OS::strncmp(str,
                       httpProtocolString_,
                       httpProtocolStringLength_) == 0) {

        str += httpProtocolStringLength_;

    }

    GetAuthorizationString(str);
    GetRawHostname(str);
    GetRequest(str);
    GetFullAddress(str);

    if(authorizationString_ != NULL)
        EncodeAuthorizationString();
}

/// @brief Get author string.
void Address::GetAuthorizationString(const ACE_TCHAR* str)
{
    const char* atCharPosition = ACE_OS::strchr(str, '@');
    if(atCharPosition != NULL) {

        const unsigned int offset = atCharPosition - str;

        for(unsigned int i = 0; i < offset; ++i)
            if(str[i] == '/')
                return;
        
        authorizationString_ = new char[ offset + 1];
        ACE_OS::memset(authorizationString_,
                       0,
                       offset + 1);
        ACE_OS::strncpy( authorizationString_,
                         str,
                         offset);
    }
}

/// @brief Get host name.
void Address::GetRawHostname(const ACE_TCHAR* str)
{
    if(authorizationString_ != NULL)
        str += ACE_OS::strlen(authorizationString_) + 1;

    const char* endPtr = str;
    while(*endPtr && *endPtr != ':' && *endPtr != '/')
        ++endPtr;

    const unsigned int offset = endPtr - str;
    rawHostName_ = new char[ offset + 1 ];
    ACE_OS::memset(rawHostName_, 0, offset + 1);
    ACE_OS::strncpy(rawHostName_, str, offset);

    if(!*endPtr || *endPtr == '/')
        return;

    // invalid character will be ignored
    portNumber_ = ACE_OS::strtoul(++endPtr, NULL, 10);
}

/// @brief Get request.
void Address::GetRequest(const ACE_TCHAR* str)
{
    const char* atSlashPosition = ACE_OS::strchr(str, '/');
    if(atSlashPosition != NULL) {

        request_ = new char[ ACE_OS::strlen(atSlashPosition) * 3 + 1];

        UrlEncode(ACE_reinterpret_cast(const unsigned char*, atSlashPosition),
                  ACE_reinterpret_cast(unsigned char*, request_));

    } else {

        request_ = new char[2];
        ACE_OS::memset(request_, 0, 2);
        ACE_OS::strncpy(request_, "/", 2);

    }
}

/// @brief Get full address.
void Address::GetFullAddress(const ACE_TCHAR* str)
{
    if(authorizationString_ != NULL)
        str += ACE_OS::strlen(authorizationString_) + 1;

    fullAddress_ = new char[ ACE_OS::strlen(str) * 3 + 
                             httpProtocolStringLength_ + 
                             1 ];
    ACE_OS::strncpy(fullAddress_,
                    httpProtocolString_,
                    httpProtocolStringLength_);

    UrlEncode(ACE_reinterpret_cast(const unsigned char*, str),
              ACE_reinterpret_cast(unsigned char*, fullAddress_ + httpProtocolStringLength_));
}

/// @brief BASE64 encode author string.
void Address::EncodeAuthorizationString()
{
    size_t encodedLength = 0;
    ACE_Byte* encodedString = ACE_Base64::encode(ACE_reinterpret_cast(ACE_Byte*, authorizationString_), ACE_OS::strlen(authorizationString_), &encodedLength);

    if(encodedString != NULL) {

        encodedAuthorizationString_ = new ACE_Byte[ encodedLength + 1];
        ACE_OS::memset(encodedAuthorizationString_, 0, encodedLength + 1);
        ACE_OS::memcpy(encodedAuthorizationString_,
                       encodedString,
                       encodedLength);
    }
}

/// @brief Urlencode function.
///
/// dest need to have enough space, maybe 3 * orig + 1.
void Address::UrlEncode(const unsigned char* orig, unsigned char* dest)
{
    while(*orig) {

        if(NeedUrlEncode(*orig)) {

            *(dest++) = '%';
            *(dest++) = DecToHex( *orig / 16 );
            *(dest++) = DecToHex( *orig % 16 );

        } else
            *(dest++) = *orig;

        ++orig;
    }
    *(dest) = '\0';
}

/// @brief Create address object.
/// @param destUrl URL string.
/// @param convertCurrentEncoding Convert string to current encoding or not.
std::auto_ptr<Address> CreateAddress(const std::string& destUrl,
                                     bool convertCurrentEncoding)
{
    if(!convertCurrentEncoding) {

        return std::auto_ptr<Address>(new Address(destUrl.c_str()));

    } else {

        try {
            std::string encoding(Glib::locale_from_utf8(destUrl));
            
            return std::auto_ptr<Address>(new Address(encoding.c_str()));

        } catch(Glib::ConvertError& err) {

            return std::auto_ptr<Address>(new Address(destUrl.c_str()));
        }
    }
}

int FileGet::open(void* p)
{
    ACE_Time_Value iterDelay(2);
    if(FileGetHandler::open(p) == -1)
        return -1;

    this->notifier_.reactor(this->reactor());
    this->msg_queue()->notification_strategy(&this->notifier_);

    if(destAddress_->Authorization() == NULL)
        NetUtil::SendGetMessage(destAddress_->FullAddress(), this->peer());
    else
        NetUtil::SendGetMessage(destAddress_->FullAddress(),
                       destAddress_->Authorization(),
                       this->peer());

    return this->reactor()->schedule_timer
        (this, 0, ACE_Time_Value::zero, iterDelay);
}

int FileGet::handle_input(ACE_HANDLE fd)
{
    char buf[1024];
    ssize_t recvCnt = this->peer().recv(buf, sizeof(buf) - 1);

    // reduce these duplicated codes :(
    if(recvCnt > 0) {

        if(!recvHeader_) {

            char* separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(buf, ' ', recvCnt));
            if(separator == NULL) {
                this->reactor()->end_reactor_event_loop();

                return -1;
            }

            switch(separator[1]) {
            case '3':
                redirect_ = true;

                break;
            case '2':

                break;
            default:
                // we interest on only success or redirect.
                this->reactor()->end_reactor_event_loop();

                return -1;
            }

            if(redirect_) {

                while(separator != NULL) {

                    if(ACE_OS::memcmp(separator + 2, "Location: ", 10) == 0) {

                        char* endLine = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 12, '\r', recvCnt - (separator - buf) - 12));
                        if(endLine != NULL) {

                            redirect_ = true;

                            const size_t len = (endLine - separator) - 11;
                            redirectURL_ = new char[len];
                            ACE_OS::memset(redirectURL_, 0, len);
                            ACE_OS::strncpy(redirectURL_, separator + 12, len - 1);

                            this->reactor()->end_reactor_event_loop();

                            return -1;
                        }
                    }

                    separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 1, '\r', recvCnt - (separator - buf) - 1));
                }

            } else {

                while(separator != NULL) {

                    if(ACE_OS::memcmp(separator, "\r\n\r\n", 4) == 0) {

                        recvHeader_ = true;
                        ACE_OS::fwrite(separator + 4,
                                       1,
                                       recvCnt - 4 - (separator - buf),
                                       outfile_);

                        break;
                    }

                    separator = ACE_reinterpret_cast(char*, ACE_OS::memchr(separator + 1, '\r', recvCnt - (separator - buf) - 1));
                }
            }

        } else {

            ACE_OS::fwrite(buf, 1, recvCnt, outfile_);

        }

        return 0;
    }

    if(recvCnt == 0 || ACE_OS::last_error() != EWOULDBLOCK) {

        this->reactor()->end_reactor_event_loop();

        return -1;
    }

    return 0;
}

int FileGet::handle_output(ACE_HANDLE fd)
{
    ACE_Message_Block* mb = NULL;
    ACE_Time_Value nowait(ACE_OS::gettimeofday());
    while(this->getq(mb, &nowait) != -1) {

        ssize_t sendCnt = this->peer().send(mb->rd_ptr(), mb->length());
        // some send error at sendCnt is -1
        if(sendCnt != -1)
            mb->rd_ptr(ACE_static_cast(size_t, sendCnt));

        if(mb->length() > 0) {

            this->ungetq(mb);
            break;
        }
        mb->release();
    }

    if(this->msg_queue()->is_empty())
        this->reactor()->cancel_wakeup
            (this, ACE_Event_Handler::WRITE_MASK);
    else
        this->reactor()->schedule_wakeup
            (this, ACE_Event_Handler::WRITE_MASK);

    return 0;
}

int FileGet::handle_timeout(const ACE_Time_Value& currentTime, const void* act)
{
    if(this->iterations_ >= ITERATIONS)
        this->peer().close_writer();

    return 0;
}
