/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player. 
//
// Orignal widget first appear in 'Air' project(http://air.kldp.net),
// and this is compact type.
//
//
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file dirbrowsewin.cc
/// @brief Class for directory browse window.

#include <vector>
#include <algorithm>

#include <sys/types.h>
#include <unistd.h>
#include <dirent.h>

#include <libintl.h>

#include "../include/dirbrowsewin.h"
#include "../include/misc.h"

std::string DirBrowseWnd::selectedDirectory_;

/// @brief Constructor of DirBrowseWnd.
/// @param border Window border width.
/// @param displayWatch Display watch in progress. If you can't realize this window, set false.
/// @param initialDirectory Directory to expand instead current directory at initial.
DirBrowseWnd::DirBrowseWnd(const int& border, const bool& displayWatch, const std::string& initialDirectory)
    : displayWatch_(displayWatch)
{
    if(!initialDirectory.empty())
        selectedDirectory_ = initialDirectory;

    set_border_width(border);
    add(treeView_);
    set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);

    treeModel_ = Gtk::TreeStore::create(column_);
    treeView_.set_model(treeModel_);
    treeView_.append_column(gettext("Path"), column_.colName_);
    treeView_.signal_row_activated().connect(sigc::mem_fun(*this, &DirBrowseWnd::OnRowActivated));
    treeView_.signal_row_expanded().connect(sigc::mem_fun(*this, &DirBrowseWnd::OnRowExpanded));

    // you need to use connect_notify !
    treeView_.signal_button_press_event().connect_notify(sigc::mem_fun(*this, &DirBrowseWnd::OnBtnPressEvent));

    Gtk::Menu_Helpers::MenuList& popupMenuList = menuPopup_.items();
    popupMenuList.push_back( Gtk::Menu_Helpers::StockMenuElem(Gtk::Stock::REFRESH, sigc::mem_fun(*this, &DirBrowseWnd::OnMenuRefresh) ));

    AddRootDirectory();
}

DirBrowseWnd::~DirBrowseWnd()
{}

/// @brief Get selected directory full path from current cursor.
const std::string& DirBrowseWnd::SelectedDirectoryFromIter()
{
    Gtk::TreeModel::Path currentPath;
    Gtk::TreeView::Column* currentColumn = NULL;

    treeView_.get_cursor(currentPath, currentColumn);

    const Gtk::TreeModel::iterator iter(treeModel_->get_iter(currentPath));
    if(iter) {

        selectedDirectory_ = (*iter)[column_.colPath_];
    }

    return selectedDirectory_;
}

/// @brief Callback function used when window is shown.
///
/// When window is shown, expand tree to current directory.
void DirBrowseWnd::on_show()
{
    Gtk::Widget::on_show();

    ExpandCurrentDir();
}

/// @brief Callback function used when button is pressed.
///
/// When button is pressed, if button is 3 button, pop up menu.
void DirBrowseWnd::OnBtnPressEvent(GdkEventButton* ev)
{
    if(ThirdButtonPressEvent(ev))
        menuPopup_.popup(ev->button, ev->time);
}

/// @brief Callback function used when row is activated.
///
/// When row is selected(double click or else ...), read a full path at row.
void DirBrowseWnd::OnRowActivated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
{
    const Gtk::TreeModel::iterator iter = treeModel_->get_iter(path);
    if(iter)
        selectedDirectory_ = (*iter)[column_.colPath_];

    signalSelected_.emit();
}

/// @brief Callback function used when row is expanded
void DirBrowseWnd::OnRowExpanded(const Gtk::TreeModel::iterator& iter, const Gtk::TreeModel::Path& path)
{
    ChangeWatchCursor();

    currentPath_ = (*iter)[column_.colPath_];
    std::vector<std::string> dirNames; 

    // If don't have sub directory, remove dummy
    if(!GetUnregisteredSubDirectories(dirNames))
        treeModel_->erase(iter->children().begin());
    else if(!dirNames.empty()) {
        std::sort(dirNames.begin(), dirNames.end());

        typedef std::vector<std::string>::const_iterator veciter;
        veciter eachFile(dirNames.begin());
        const veciter vecEnd(dirNames.end());

        Gtk::TreeModel::Row child( *(iter->children().begin()) );
        InsertName(child, *eachFile);
        ++eachFile;

        while(eachFile != vecEnd) {
            child = *(treeModel_->append(iter->children()));
            InsertName(child, *eachFile);
            ++eachFile;
        }
    }

    ChangeMainCursor();
}

/// @brief Expand tree to current(initial) directory.
void DirBrowseWnd::ExpandCurrentDir()
{
    if(selectedDirectory_.empty() || !IsExecutable(selectedDirectory_) || !IsReadable(selectedDirectory_))
        selectedDirectory_ = Glib::get_current_dir();

    treeView_.expand_row(Gtk::TreePath("0"), false);
    Gtk::TreeNodeChildren child(treeModel_->get_iter("0")->children());
    typedef Gtk::TreeNodeChildren::iterator childiter;
    childiter tempParent;

    std::string::const_iterator pathBegin(selectedDirectory_.begin());
    std::string::const_iterator temp;
    const std::string::const_iterator pathEnd(selectedDirectory_.end());

    Glib::ustring compareName;
    while(pathBegin != pathEnd) {

        pathBegin = find_if(pathBegin, pathEnd, std::not1(IsCharSlash()));
        temp = find_if(pathBegin, pathEnd, IsCharSlash());
        if(pathBegin != pathEnd) {

            try {

                compareName = Glib::locale_to_utf8(std::string(pathBegin, temp));
            } catch(Glib::ConvertError& err) {

                compareName = std::string(pathBegin, temp);

            }

            for(childiter iter = child.begin(); iter != child.end(); ++iter) {

                if(iter->get_value(column_.colName_) == compareName) {
                    tempParent = iter;
                    treeView_.expand_row(Gtk::TreePath(iter), false);
                }

            }

            child = tempParent->children();
        }

        pathBegin = temp;
    }

    // Scroll to current directory to mid
    treeView_.scroll_to_row(Gtk::TreePath(tempParent), 0.5);
    treeView_.set_cursor(Gtk::TreePath(tempParent));
}


/// @brief Callback funtion used when refresh menu is selected.
///
/// Swap external container to empty one and expand tree to current(initial) directory.
void DirBrowseWnd::OnMenuRefresh()
{
    Gtk::TreeModel::iterator iter(treeModel_->children().begin());

    while(iter)
        iter = treeModel_->erase(iter);

    ResetPathSet();

    AddRootDirectory();

    ExpandCurrentDir();
}

/// @brief Get unregistered sub directoy list.
/// @param dest Result container.
/// @return Retrun true if directory has any sub directory.
bool DirBrowseWnd::GetUnregisteredSubDirectories(std::vector<std::string>& dest)
{
    bool rtn = false;
    DIR* const currentDir = opendir(currentPath_.c_str());
    dirent* dirEntry = NULL;
    while((dirEntry = readdir(currentDir)) != NULL) {
        if(strncmp(dirEntry->d_name, ".", 1) && strncmp(dirEntry->d_name, "..", 2)) {
            fullPath_ = currentPath_;
            fullPath_ += '/';
            fullPath_ += dirEntry->d_name;
            if(IsDirectory(fullPath_)) {
                rtn = true;
                if(!pathInStore_.count(fullPath_))
                    dest.push_back(dirEntry->d_name);
            }
        }
    }
    closedir(currentDir); 

    return rtn;
}

void DirBrowseWnd::InsertName(Gtk::TreeModel::Row& child, const std::string& name)
{
    GetFullPathname(name);
    pathInStore_.insert(fullPath_);
    FillRow(child, name);
}
