// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file effect.cc
/// @brief Effector functions source file.

#include "../include/effect.h"

namespace Effect {
   
    Base BassBoost_L[2], BassBoost_R[2];
    Base BassBoost2_L[3], BassBoost2_R[3];
    Base TrebleBoost_L, TrebleBoost_R;
    Base GraphicEQ_L[11], GraphicEQ_R[11];
    Base Room_L, Room_R;
    Base Room2_L[6], Room2_R[6];
    Base Harmony_L, Harmony_R;

    Peak Bass40, Bass2_80;;
    
    Peak EQ50, EQ100, EQ200, EQ400, EQ800, EQ1500, EQ3k, EQ5k;
    Peak EQ7k, EQ10k, EQ15k;
    
    LowShelf Bass100, Bass2_199;
    LowShelf Room2_35, Room2_62, Room2_279, Room2_852;
    HighShelf Treble5k, Bass2_1309;
    HighShelf Room2_3713, Room2_13545; 
    LowPass LPF_;
    HighPass HPF_;

    double prevLeftSample[5];
    double prevRightSample[5];
    
    double prevLeftSample1[3];
    double prevRightSample1[3];

    int room2Count[2];

    double room2PrevSample[1764];
    double prevAvg[2];

    int prevCount;
    int prevRoomCount;
    double room2CrossAmplifyMount;
    int prevHarmonyCount;
    
    double peakSample;
    double feedGain;
    double feedRhythme;
    double amplifyRatio;

    bool needCalc;
    
}
/// @brief Apply clipping.
///
/// If sample over peakSample, calculate feedGain to reduce
/// global gain.
/// @param sample Sample to apply.
void Effect::Clip(double& sample)
{
    if(sample > 1.0) {
	if(sample > peakSample) {
	    peakSample = sample;

	    feedGain = pow(10.0, -log((peakSample - 1.0) * 0.5 + 1.0));
	    
	}
	sample = 1.0;
	
    } else if(sample < -1.0) {
	if(sample < -peakSample) {
	    peakSample = -sample;
	    
	    feedGain = pow(10.0, -log((peakSample - 1.0) * 0.5 + 1.0));

	}
	sample = -1.0;

    }
}

/// @brief Apply clipping to stereo.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
/// @sa Effect::Clip(double& sampleL, double& sampleR)
inline void Effect::Clip(double& sampleL, double& sampleR)
{
    Clip(sampleL);
    Clip(sampleR);
}

/// @brief Amplify sample.
///
/// @param sampleL Left sample to amplify.
/// @param sampleR Right sample to amplify.
/// @param gain Gain.
inline void Effect::Amplify(double& sampleL, double& sampleR, const float& gain)
{
    sampleL *= gain;
    sampleR *= gain;
}

/// @brief Amplify sample and apply clipping.
///
/// @param sampleL Left sample to amplify.
/// @param sampleR Right sample to amplify.
/// @param gain Gain.
inline void Effect::AmplifyClip(double& sampleL, double& sampleR, const float& gain)
{
    Amplify(sampleL, sampleR, gain);

    Clip(sampleL, sampleR);
}

/// @brief Amplify sample and apply clipping.
///
/// @param sampleL Left sample to amplify.
/// @param sampleR Right sample to amplify.
/// @param gain Gain in decibel.
inline void Effect::AmplifyDecibel(double& sampleL, double& sampleR, const float& gain)
{ 
    amplifyRatio = pow(10.0, gain * 0.05);

    AmplifyClip(sampleL, sampleR, amplifyRatio);
}

/// @brief Apply stereo enhance effector.
///
/// Add difference of stereo signal.
/// @param left Left sample to apply.
/// @param right Right sample to apply.
/// @param wide mount of difference. 0.1 ~ 2.0 
void Effect::StereoEnhance(double& left, double& right, const float& wide)
{
    const double stereoEnhanceDiffLeft = (left - right) * 0.5 * wide;
    
    left += stereoEnhanceDiffLeft;
    right -= stereoEnhanceDiffLeft;

    Clip(left, right);

}

/// @brief Apply bass boost effector.
///
/// Apply low shelf EQ and peak EQ at low frequency.
/// @param sampleRate Sample rate.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
void Effect::BassBoost(const int& sampleRate, double& sampleL, double& sampleR)
{
    if(Bass100.SampleRate() != sampleRate)
    {	
	Bass100.Calc(sampleRate, 100, 3.0f, 0.60295);
    }

    Bass100.Apply(sampleL, BassBoost_L[0]);
    Bass100.Apply(sampleR, BassBoost_R[0]);

    Clip(sampleL, sampleR);

    EQ(Effect::Bass40, sampleRate, sampleL, sampleR, 40, 9.0f, BassBoost_L[1], BassBoost_R[1], 1.4f);

}

/// @brief Apply bass boost2 effector.
///
/// Apply low shelf EQ, high shelf EQ and peak EQ.
/// @param sampleRate Sample rate.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
void Effect::BassBoost2(const int& sampleRate, double& sampleL, double& sampleR)
{
    if(Bass2_80.SampleRate() != sampleRate) {
	
	Bass2_80.Calc(sampleRate, 80, 6.5f, 0.76077);
	Bass2_199.Calc(sampleRate, 199, 5.5f, 0.33374);
	Bass2_1309.Calc(sampleRate, 1309, 9.0f, 0.25842);

    }

    Bass2_199.Apply(sampleL, BassBoost2_L[0]);
    Bass2_199.Apply(sampleR, BassBoost2_R[0]);
    
    Bass2_1309.Apply(sampleL, BassBoost2_L[1]);
    Bass2_1309.Apply(sampleR, BassBoost2_R[1]);
    
    Bass2_80.Apply(sampleL, BassBoost2_L[2]);
    Bass2_80.Apply(sampleR, BassBoost2_R[2]);
    
    Clip(sampleL, sampleR);

}

/// @brief Apply treble boost effector.
///
/// Apply high shelf EQ at high frequency. When boost high frequency,
/// high frequency noise increase also.
/// @param sampleRate Sample rate.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
void Effect::TrebleBoost(const int& sampleRate, double& sampleL, double& sampleR)
{
    if(Treble5k.SampleRate() != sampleRate)
    {	
	Treble5k.Calc(sampleRate, 5000, 6.0f, 0.7);
    }

    Treble5k.Apply(sampleL, TrebleBoost_L);
    Treble5k.Apply(sampleR, TrebleBoost_R);

    Clip(sampleL, sampleR);

}

/// @brief Apply peak EQ effector.
void Effect::EQ(BiquadCoeff& peak, const int& sampleRate, double& sampleL, double& sampleR, const int& freq, const float& gain, Base& prevL, Base& prevR, const float& Q)
{
    if(peak.SampleRate() != sampleRate)
    {
	peak.Calc(sampleRate, freq, gain, Q);
    }
    
    peak.Apply(sampleL, prevL);
    peak.Apply(sampleR, prevR);

    Clip(sampleL, sampleR);

}

/// @brief Apply 11 band graphic EQ effector.
///
/// Apply peak EQ to 50, 100, 200, 400, 800, 1500, 3k, 5k, 7k, 10k, 15k frequency.
/// @param sampleRate Sample rate.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
/// @param parameter Float array of gain value. Array size is 11.
/// @param calc Need to calculate.
void Effect::GraphicEQ(const int& sampleRate, double& sampleL, double& sampleR, const float* const parameter, const bool& calc)
{
    if(calc || EQ50.SampleRate() != sampleRate) {
	EQ50.Calc(sampleRate, 50, parameter[0], 1.0);
	EQ100.Calc(sampleRate, 100, parameter[1], 1.0);
	EQ200.Calc(sampleRate, 200, parameter[2], 1.0);
	EQ400.Calc(sampleRate, 400, parameter[3], 1.0);
	EQ800.Calc(sampleRate, 800, parameter[4], 1.0);
	EQ1500.Calc(sampleRate, 1500, parameter[5], 1.0);
	EQ3k.Calc(sampleRate, 3000, parameter[6], 1.0);
	EQ5k.Calc(sampleRate, 5000, parameter[7], 1.0);
	EQ7k.Calc(sampleRate, 7000, parameter[8], 1.0);
	EQ10k.Calc(sampleRate, 10000, parameter[9], 1.0);
	EQ15k.Calc(sampleRate, 15000, parameter[10], 1.0);
    }

    EQ50.Apply(sampleL, GraphicEQ_L[0]);
    EQ50.Apply(sampleR, GraphicEQ_R[0]);
    
    EQ100.Apply(sampleL, GraphicEQ_L[1]);
    EQ100.Apply(sampleR, GraphicEQ_R[1]);

    EQ200.Apply(sampleL, GraphicEQ_L[2]);
    EQ200.Apply(sampleR, GraphicEQ_R[2]);

    EQ400.Apply(sampleL, GraphicEQ_L[3]);
    EQ400.Apply(sampleR, GraphicEQ_R[3]);

    EQ800.Apply(sampleL, GraphicEQ_L[4]);
    EQ800.Apply(sampleR, GraphicEQ_R[4]);

    EQ1500.Apply(sampleL, GraphicEQ_L[5]);
    EQ1500.Apply(sampleR, GraphicEQ_R[5]);

    EQ3k.Apply(sampleL, GraphicEQ_L[6]);
    EQ3k.Apply(sampleR, GraphicEQ_R[6]);

    EQ5k.Apply(sampleL, GraphicEQ_L[7]);
    EQ5k.Apply(sampleR, GraphicEQ_R[7]);
    
    EQ7k.Apply(sampleL, GraphicEQ_L[8]);
    EQ7k.Apply(sampleR, GraphicEQ_R[8]);

    EQ10k.Apply(sampleL, GraphicEQ_L[9]);
    EQ10k.Apply(sampleR, GraphicEQ_R[9]);
    
    EQ15k.Apply(sampleL, GraphicEQ_L[10]);
    EQ15k.Apply(sampleR, GraphicEQ_R[10]);

    Clip(sampleL, sampleR);
}

/// @brief Apply low pass filter.
Effect::Stereo Effect::LPF(BiquadCoeff& p, const int& sampleRate, double& sampleL, double& sampleR, const int& freq, const double& gain, Base& prevL, Base& prevR)
{
    if(p.SampleRate() != sampleRate)
    {
	p.Calc(sampleRate, freq, gain, 0.0);
    }

    return Stereo( p.Return(sampleL, prevL), p.Return(sampleR, prevR) );
}

/// @brief Apply high pass filter.
Effect::Stereo Effect::HPF(BiquadCoeff& p, const int& sampleRate, double& sampleL, double& sampleR, const int& freq, const double& gain, Base& prevL, Base& prevR)
{
    if(p.SampleRate() != sampleRate)
    {
	p.Calc(sampleRate, freq, gain, 0.0);
    }

    return Stereo( p.Return(sampleL, prevL), p.Return(sampleR, prevR) );
}

/// @brief Apply rhythem effector.
///
/// Adjust gain according to average of stereo signal.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
/// @param mount Limit of adjustable gain.
void Effect::Rhythme(double& sampleL, double& sampleR, const double& mount)
{
    if(prevCount != 18) {
	prevCount++;
	return;
    } else { 
	// if prevCount() == 18
	const double avg = sqrt(sampleL * sampleR);

	if(feedRhythme < mount && feedRhythme > -mount + 0.2 && prevAvg[0] < avg) { 
	    if(prevAvg[0] * avg < 0.0)
		feedRhythme = 0.0;
	    if(prevAvg[1] < prevAvg[0])
		// increase about 0.2 dB
		feedRhythme += 0.023293;
	    else 
		// increase about 0.1 dB
		feedRhythme += 0.0115795;
	} else if(feedRhythme < mount - 0.2 && feedRhythme > -mount && prevAvg[0] > avg) { 
	    if(prevAvg[0] * avg < 0.0)
		feedRhythme = 0.0;
	    if(prevAvg[1] > prevAvg[0])
		//decrease about 0.2 dB
		feedRhythme -= 0.0227628;
	    else
		// decrease about 0.1 dB  
		feedRhythme -= 0.0114469;
	}
	
	prevCount = 0;
	prevAvg[1] = prevAvg[0];
	prevAvg[0] = avg;

    }

    // There is no change in signal this time, there is no clipping.
}

/// @brief Apply room effector.
///
/// Add previous signal that LPF processed.
/// @param sampleRate Sample rate.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
/// @param ab Feed mount of previous signal.
void Effect::Room(const int& sampleRate, double& sampleL, double& sampleR, const double& ab)
{
    Stereo filtered = LPF(LPF_, sampleRate, sampleL, sampleR, 5300, 0.75, Room_L, Room_R); 

    const double abSquare = ab * ab;

    if(prevRoomCount > 4)
	prevRoomCount = 0;

    sampleL += prevLeftSample[0] * ab * 0.75 + prevLeftSample[1] * ab * 0.45	    
	    + prevLeftSample[2] * abSquare * 0.60 + prevLeftSample[3] * ab * 0.15
	    + prevLeftSample[4] * abSquare * 0.30;
	    
    sampleR += prevRightSample[0] * ab * 0.75 + prevRightSample[1] * ab * 0.45	    
	    + prevRightSample[2] * abSquare * 0.60 + prevRightSample[3] * ab * 0.15
	    + prevRightSample[4] * abSquare * 0.30;

    prevLeftSample[prevRoomCount] = filtered.left;
    prevRightSample[prevRoomCount] = filtered.right;

    prevRoomCount++;

    Clip(sampleL, sampleR);

}

/// @brief Apply room2 effector.
///
/// Add previous signal and apply EQ.
/// @param sampleRate Sample rate.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
/// @param mount Shape of EQ.
/// @param calc Need to calculate.
void Effect::Room2(const int& sampleRate, double& sampleL, double& sampleR, const double& mount, const bool& calc)
{
    if(Room2_62.SampleRate() != sampleRate) {

	Room2_62.Calc(sampleRate, 62, -8.0f, 0.51208);
	Room2_279.Calc(sampleRate, 279, 4.0f, 0.39539);
	Room2_13545.Calc(sampleRate, 13545, 4.0f, 0.37929);

	Room2_35.Calc(sampleRate, 35, -3.0f, 0.52443);
	Room2_852.Calc(sampleRate, 852, 8.0f + (10.0 * mount), 0.37624);
	Room2_3713.Calc(sampleRate, 3713, 12.0f, 0.39159);

    }

    if(calc) {
	Room2_852.Calc(sampleRate, 852, 8.0f + (10.0 * mount), 0.37624);

	// NOTES. i do not reset room2CrossAmplifyMount, so be careful
	// waste value in room2CrossAmplifyMount.
	room2CrossAmplifyMount = pow(10.0, (mount * -10.0 - 15.0) * 0.05);
	 
    }

    if(room2Count[0] >= 882)
	room2Count[0] = 0;

    room2Count[1] = room2Count[0];
    room2Count[1] -= 441;
    
    if(room2Count[1] < 0)
	room2Count[1] += 882;
	
    sampleL += room2PrevSample[ room2Count[1] ] * 0.146997;
    room2PrevSample[ room2Count[0] ] = sampleL;		
    
    sampleR += room2PrevSample[ room2Count[1] + 882 ] * 0.146997;
    room2PrevSample[ room2Count[0] + 882 ] = sampleR;		

    ++room2Count[0];

    Room2_279.Apply(sampleL, Room2_L[0]);
    Room2_279.Apply(sampleL, Room2_R[0]);

    Room2_62.Apply(sampleL, Room2_L[1]);
    Room2_62.Apply(sampleR, Room2_R[1]);

    Room2_13545.Apply(sampleL, Room2_L[2]);
    Room2_13545.Apply(sampleR, Room2_R[2]);

    AmplifyClip(sampleL, sampleR, 0.70795);

    Stereo cross(sampleR, sampleL);
    
    Room2_852.Apply(cross.left, Room2_L[3]);
    Room2_852.Apply(cross.right, Room2_R[3]);

    Room2_35.Apply(cross.left, Room2_L[4]);
    Room2_35.Apply(cross.right, Room2_R[4]);

    Room2_3713.Apply(cross.left, Room2_L[5]);
    Room2_3713.Apply(cross.right, Room2_R[5]);

    Amplify(cross.left, cross.right, room2CrossAmplifyMount);

    sampleL += cross.left;
    sampleR += cross.right;

    Clip(sampleL, sampleR);

}

/// @brief Apply harmony effector.
///
/// Add previous signal that processed HPF on high average power.
/// @param sampleRate Sample rate.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
/// @param ab Feed mount of previous signal.
void Effect::Harmony(const int& sampleRate, double& sampleL, double& sampleR, const double& ab)
{
    Stereo filtered = HPF(HPF_, sampleRate, sampleL, sampleR, 10000, 0.5, Harmony_L, Harmony_R); 
    
    if(prevHarmonyCount > 2)
	prevHarmonyCount = 0;

    prevLeftSample1[prevHarmonyCount] = filtered.left;
    prevRightSample1[prevHarmonyCount] = filtered.right;
    
    prevHarmonyCount++;

    const double sampleMultiple = sampleL * sampleR;

    if(sampleMultiple > 0.5625 || sampleMultiple < -0.5625) {
	const double abSquare = ab * ab;

	sampleL += prevLeftSample1[2] * abSquare * 0.1 + prevLeftSample1[1] * ab * 0.3
		+ prevLeftSample1[0] * ab * 0.5;

	sampleR += prevRightSample1[2] * abSquare * 0.1 + prevRightSample1[1] * ab * 0.3
		+ prevRightSample1[0] * ab * 0.5;
		
	Clip(sampleL, sampleR);
    }

}

/// @brief Apply distortion effector.
/// @param sampleL Left sample to apply.
/// @param sampleR Right sample to apply.
/// @param mount Mount of distortion. Higher value produce hard sample. 
void Effect::Distortion(double& sampleL, double& sampleR, const double& mount)
{
    if(mount > 5.0) {
	sampleL = 0.2 * atan(sampleL * mount);
	
	sampleR = 0.2 * atan(sampleR * mount);
    
    } else {
	const double multiple = 1.0 / mount;
	
	sampleL = multiple * atan(sampleL * mount);
	
	sampleR = multiple * atan(sampleR * mount);

    }
}

/// @brief Process stereo signal.
///
/// Read preference and apply effector to signal.
/// @param preference Preference.
/// @param sampleRate Sample rate.
/// @param orig Input stereo signal to apply.
void Effect::Process(const Preference& preference, const int& sampleRate, Stereo& orig)
{
    needCalc = false;
    if(preference.EfxChanged()) {
	needCalc = true;
	preference.EfxChangedUnset();
    }
    
    AmplifyClip(orig.left, orig.right, feedGain + feedRhythme);

    if(preference.EfxBassBoost()) 
	BassBoost(sampleRate, orig.left, orig.right);

    if(preference.EfxBassBoost2())
	BassBoost2(sampleRate, orig.left, orig.right);

    if(preference.EfxTrebleBoost())
	TrebleBoost(sampleRate, orig.left, orig.right);
 
    if(preference.EfxGraphicEQ())
	GraphicEQ(sampleRate, orig.left, orig.right, preference.EfxGraphicEQGain(), needCalc);	
 
    if(preference.EfxStereoEnhance())
	StereoEnhance(orig.left, orig.right, preference.EfxStereoEnhanceAmount());

    if(preference.EfxRhythme())
	Rhythme(orig.left, orig.right, preference.EfxRhythmeAmount());
  
    if(preference.EfxRoom())
	Room(sampleRate, orig.left, orig.right, preference.EfxRoomAmount());

    if(preference.EfxRoom2())
	Room2(sampleRate, orig.left, orig.right, preference.EfxRoom2Amount(), needCalc);

    if(preference.EfxHarmony())
	Harmony(sampleRate, orig.left, orig.right, preference.EfxHarmonyAmount());

    if(preference.EfxDistortion())
	Distortion(orig.left, orig.right, preference.EfxDistortionAmount());
    
}

/// @brief Reset static variables.
void Effect::Reset()
{
    peakSample = 0.0;
    feedGain = 1.0;
    feedRhythme = 0.0;
    prevAvg[0] = 0.0;
    prevAvg[1] = 0.0;
    prevCount = 0;
    prevRoomCount = 0;
    prevHarmonyCount = 0;

    prevLeftSample[0] = 0.0;
    prevLeftSample[1] = 0.0;
    prevLeftSample[2] = 0.0;
    prevLeftSample[3] = 0.0;
    prevLeftSample[4] = 0.0;
    
    prevRightSample[0] = 0.0;
    prevRightSample[1] = 0.0;
    prevRightSample[2] = 0.0;
    prevRightSample[3] = 0.0;
    prevRightSample[4] = 0.0;

    prevLeftSample1[0] = 0.0;
    prevLeftSample1[1] = 0.0;
    prevLeftSample1[2] = 0.0;
    prevRightSample1[0] = 0.0;
    prevRightSample1[1] = 0.0;
    prevRightSample1[2] = 0.0;

    room2Count[0] = 0;
    room2Count[1] = 0;

    for(unsigned int i = 0; i < 1764; ++i)
    {
	room2PrevSample[i] = 0.0;
    }

    amplifyRatio = 0.0;

    BassBoost_L[0].Reset();
    BassBoost_L[1].Reset();
    BassBoost_R[0].Reset();
    BassBoost_R[1].Reset();

    BassBoost2_L[0].Reset();
    BassBoost2_L[1].Reset();
    BassBoost2_L[2].Reset();
    BassBoost2_R[0].Reset();
    BassBoost2_R[1].Reset();
    BassBoost2_R[2].Reset();

    TrebleBoost_L.Reset();
    TrebleBoost_R.Reset();

    GraphicEQ_L[0].Reset();
    GraphicEQ_L[1].Reset();
    GraphicEQ_L[2].Reset();
    GraphicEQ_L[3].Reset();
    GraphicEQ_L[4].Reset();
    GraphicEQ_L[5].Reset();
    GraphicEQ_L[6].Reset();
    GraphicEQ_L[7].Reset();
    GraphicEQ_L[8].Reset();
    GraphicEQ_L[9].Reset();
    GraphicEQ_L[10].Reset();
    
    GraphicEQ_R[0].Reset();
    GraphicEQ_R[1].Reset();
    GraphicEQ_R[2].Reset();
    GraphicEQ_R[3].Reset();
    GraphicEQ_R[4].Reset();
    GraphicEQ_R[5].Reset();
    GraphicEQ_R[6].Reset();
    GraphicEQ_R[7].Reset();
    GraphicEQ_R[8].Reset();
    GraphicEQ_R[9].Reset();
    GraphicEQ_R[10].Reset();

    Room_L.Reset();
    Room_R.Reset();

    Room2_L[0].Reset();
    Room2_L[1].Reset();
    Room2_L[2].Reset();
    Room2_L[3].Reset();
    Room2_L[4].Reset();
    Room2_L[5].Reset();
    Room2_R[0].Reset();
    Room2_R[1].Reset();
    Room2_R[2].Reset();
    Room2_R[3].Reset();
    Room2_R[4].Reset();
    Room2_R[5].Reset();

    Harmony_L.Reset();
    Harmony_R.Reset();

}

