// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// you know, it from 'Modern C++ Design' by Andrei.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file fixedallocator.cc
/// @brief Class for fixed size allocator source file.
///
/// Fixed size allocator from 'Modern C++ Design' by Andrei.

#include <cassert>
#include <algorithm>

#include "../include/fixedallocator.h"

MySmallObjAllocator& SmallObjAllocatorInstance()
{
    static MySmallObjAllocator allocator;

    return allocator;
}

void FixedAllocator::Chunk::Init(std::size_t blockSize, unsigned char blocks)
{
    assert(blockSize > 0);
    assert(blocks > 0);
    
    assert((blockSize * blocks) / blockSize == blocks);

    pData_ = new unsigned char[blockSize * blocks];
    Reset(blockSize, blocks);
}

void FixedAllocator::Chunk::Reset(std::size_t blockSize, unsigned char blocks)
{
    assert(blockSize > 0);
    assert(blocks > 0);
    
    assert((blockSize * blocks) / blockSize == blocks);

    firstAvailableBlock_ = 0;
    blocksAvailable_ = blocks;

    unsigned char i = 0;
    unsigned char* p = pData_;
    for (; i != blocks; p += blockSize)
    {
	*p = ++i;
    }
}

void FixedAllocator::Chunk::Release()
{
    delete[] pData_;
}

void* FixedAllocator::Chunk::Allocate(std::size_t blockSize)
{
    if (!blocksAvailable_) return 0;

    assert((firstAvailableBlock_ * blockSize) / blockSize == 
	    firstAvailableBlock_);

    unsigned char* pResult =
	pData_ + (firstAvailableBlock_ * blockSize);
    firstAvailableBlock_ = *pResult;
    --blocksAvailable_;

    return pResult;
}

void FixedAllocator::Chunk::Deallocate(void* p, std::size_t blockSize)
{
    assert(p >= pData_);

    unsigned char* toRelease = static_cast<unsigned char*>(p);
    
    assert((toRelease - pData_) % blockSize == 0);

    *toRelease = firstAvailableBlock_;
    firstAvailableBlock_ = static_cast<unsigned char>(
	    (toRelease - pData_) / blockSize);
    
    assert(firstAvailableBlock_ == (toRelease - pData_) / blockSize);

    ++blocksAvailable_;
}

FixedAllocator::FixedAllocator(std::size_t blockSize)
    : blockSize_(blockSize),
    allocChunk_(0),
    deallocChunk_(0)
{
    assert(blockSize_ > 0);

    prev_ = next_ = this;

    std::size_t numBlocks = DEFAULT_CHUNK_SIZE / blockSize;
    if (numBlocks > UCHAR_MAX) numBlocks = UCHAR_MAX;
    else if (numBlocks == 0) numBlocks = 8 * blockSize;

    numBlocks_ = static_cast<unsigned char>(numBlocks);
    assert(numBlocks_ == numBlocks);
}

FixedAllocator::FixedAllocator(const FixedAllocator& rhs)
    : blockSize_(rhs.blockSize_),
    numBlocks_(rhs.numBlocks_),
    chunks_(rhs.chunks_)
{
    prev_ = &rhs;
    next_ = rhs.next_;
    rhs.next_->prev_ = this;
    rhs.next_ = this;

    allocChunk_ = rhs.allocChunk_
	? &chunks_.front() + (rhs.allocChunk_ - &rhs.chunks_.front())
	: 0;

    deallocChunk_ = rhs.deallocChunk_
	? &chunks_.front() + (rhs.deallocChunk_ - &rhs.chunks_.front())
	: 0;
}

FixedAllocator& FixedAllocator::operator=(const FixedAllocator& rhs)
{
    FixedAllocator copy(rhs);
    copy.Swap(*this);
    return *this;
}

FixedAllocator::~FixedAllocator()
{
    if (prev_ != this)
    {
	prev_->next_ = next_;
	next_->prev_ = prev_;
	return;
    }

    assert(prev_ == next_);
    Chunks::iterator i = chunks_.begin();
    for (; i != chunks_.end(); ++i)
    {
	assert(i->blocksAvailable_ == numBlocks_);
	i->Release();
    }
}

void FixedAllocator::Swap(FixedAllocator& rhs)
{
    using namespace std;

    swap(blockSize_, rhs.blockSize_);
    swap(numBlocks_, rhs.numBlocks_);
    chunks_.swap(rhs.chunks_);
    swap(allocChunk_, rhs.allocChunk_);
    swap(deallocChunk_, rhs.deallocChunk_);
}

void* FixedAllocator::Allocate()
{
    if (allocChunk_ == 0 || allocChunk_->blocksAvailable_ == 0)
    {
	Chunks::iterator i = chunks_.begin();
	for (;; ++i)
	{
	    if (i == chunks_.end())
	    {
		chunks_.reserve(chunks_.size() + 1);
		Chunk newChunk;
		newChunk.Init(blockSize_, numBlocks_);
		chunks_.push_back(newChunk);
		allocChunk_ = &chunks_.back();
		deallocChunk_ = &chunks_.front();
		break;
	    }
	    if (i->blocksAvailable_ > 0)
	    {
		allocChunk_ = &*i;
		break;
	    }
	}
    }
    assert(allocChunk_ != 0);
    assert(allocChunk_->blocksAvailable_ > 0);

    return allocChunk_->Allocate(blockSize_);
}

void FixedAllocator::Deallocate(void* p)
{
    assert(!chunks_.empty());
    assert(&chunks_.front() <= deallocChunk_);
    assert(&chunks_.back() >= deallocChunk_);

    deallocChunk_  = VicinityFind(p);
    assert(deallocChunk_);

    DoDeallocate(p);
}

FixedAllocator::Chunk* FixedAllocator::VicinityFind(void* p)
{
    assert(!chunks_.empty());
    assert(deallocChunk_);

    const std::size_t chunkLength = numBlocks_ * blockSize_;

    Chunk* lo = deallocChunk_;
    Chunk* hi = deallocChunk_ + 1;
    Chunk* loBound = &chunks_.front();
    Chunk* hiBound = &chunks_.back() + 1;

    for (;;)
    {
	if (lo)
	{
	    if (p >= lo->pData_ && p < lo->pData_ + chunkLength)
	    {
		return lo;
	    }
	    if (lo == loBound) lo = 0;
	    else --lo;
	}

	if (hi)
	{
	    if (p >= hi->pData_ && p < hi->pData_ + chunkLength)
	    {
		return hi;
	    }
	    if (++hi == hiBound) hi = 0;
	}
    }
    assert(false);
    return 0;
}

void FixedAllocator::DoDeallocate(void* p)
{
    assert(deallocChunk_->pData_ <= p);
    assert(deallocChunk_->pData_ + numBlocks_ * blockSize_ > p);

    deallocChunk_->Deallocate(p, blockSize_);

    if (deallocChunk_->blocksAvailable_ == numBlocks_)
    {

	Chunk& lastChunk = chunks_.back();

	if (&lastChunk == deallocChunk_)
	{

	    if (chunks_.size() > 1 && 
		    deallocChunk_[-1].blocksAvailable_ == numBlocks_)
	    {
		lastChunk.Release();
		chunks_.pop_back();
		allocChunk_ = deallocChunk_ = &chunks_.front();
	    }
	    return;
	}

	if (lastChunk.blocksAvailable_ == numBlocks_)
	{
	    lastChunk.Release();
	    chunks_.pop_back();
	    allocChunk_ = deallocChunk_;
	}
	else
	{
	    std::swap(*deallocChunk_, lastChunk);
	    allocChunk_ = &chunks_.back();
	}
    }
}

SmallObjAllocator::SmallObjAllocator(
	std::size_t chunkSize, 
	std::size_t maxObjectSize)
    : pLastAlloc_(0), pLastDealloc_(0)
, chunkSize_(chunkSize), maxObjectSize_(maxObjectSize) 
{   
}

void* SmallObjAllocator::Allocate(std::size_t numBytes)
{
    if (numBytes > maxObjectSize_) return operator new(numBytes);

    if (pLastAlloc_ && pLastAlloc_->BlockSize() == numBytes)
    {
	return pLastAlloc_->Allocate();
    }
    Pool::iterator i = std::lower_bound(pool_.begin(), pool_.end(), numBytes);
    if (i == pool_.end() || i->BlockSize() != numBytes)
    {
	i = pool_.insert(i, FixedAllocator(numBytes));
	pLastDealloc_ = &*pool_.begin();
    }
    pLastAlloc_ = &*i;
    return pLastAlloc_->Allocate();
}

void SmallObjAllocator::Deallocate(void* p, std::size_t numBytes)
{
    if (numBytes > maxObjectSize_) return operator delete(p);

    if (pLastDealloc_ && pLastDealloc_->BlockSize() == numBytes)
    {
	pLastDealloc_->Deallocate(p);
	return;
    }
    Pool::iterator i = std::lower_bound(pool_.begin(), pool_.end(), numBytes);
    assert(i != pool_.end());
    assert(i->BlockSize() == numBytes);
    pLastDealloc_ = &*i;
    pLastDealloc_->Deallocate(p);
}

