/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file mainmanager.cc
/// @brief Class for main program control mediator source file.

#include "../include/mainmanager.h"
#include "../include/aboutdlg.h"

/// @brief Constructor of MainManager.
/// @param arguments Command line arguments.
/// @param socketHandler UnixSocketHandler object to accept a connection.
MainManager::MainManager(const std::vector<std::string>& arguments, UnixSocketHandler& socketHandler)
    : audioFormatLittleEndian_(AO_FMT_LITTLE),
      audioFormatBigEndian_(AO_FMT_BIG),
      audioFormatNativeEndian_(AO_FMT_NATIVE),
      programDir_(Glib::get_home_dir() + "/.sagua"),
      isCharSetUTF8_(Glib::get_charset(currentCharSet_)),
      socketHandler_(socketHandler),
      trayIcon_(*this),
      mixer_(Mixer::Create(*this)),
      mainWnd_(*this, arguments)
{
    audio_ = Audio::Create(*this);

    trayIconFirstBtnCallbacks_["fileinfo"] = &MainManager::FileInfo;
    trayIconFirstBtnCallbacks_["next"] = &MainManager::PlayCtrlNext;
    trayIconFirstBtnCallbacks_["openwin"] = &MainManager::MainWndOpen;
    trayIconFirstBtnCallbacks_["pause"] = &MainManager::AudioPauseToggle;
    trayIconFirstBtnCallbacks_["play"] = &MainManager::PlayCtrlPlay;
    trayIconFirstBtnCallbacks_["prev"] = &MainManager::PlayCtrlPrev;
    trayIconFirstBtnCallbacks_["stop"] = &MainManager::PlayCtrlStop;

    signalAudioDeviceOpenFailed_.connect(sigc::mem_fun(mainWnd_, &MainWnd::OnAudioDeviceOpenFailed));

    mainWnd_.show();

    if(preference_.SuccessLoadConfig())
        mainWnd_.PlayQueue();
    else
        mainWnd_.OnMenuPreference();
}

MainManager::~MainManager()
{}

/// @brief Show MainWnd.
void MainManager::MainWndOpen()
{
    mainWnd_.show();

    mainWnd_.raise();
}

/// @brief Show about dialog.
void MainManager::About()
{
    AboutDlg aboutDlg;

    aboutDlg.run();
}

/// @brief Hide MainWnd and exit loop.
void MainManager::Quit()
{
    mainWnd_.SaveState();
    mainWnd_.hide();

    Gtk::Main::quit();
}

Glib::ustring MainManager::StringToUTF8(const Glib::ustring& orig, const std::string& fail) const
{
    if(orig.validate())
        return orig;

    if(isCharSetUTF8_)
        return StringToUTF8WithFail(orig, fail, GetAssumeCharset());
    else
        return StringToUTF8WithFail(orig, fail, currentCharSet_);
}

Glib::ustring MainManager::StringToUTF8(const Glib::ustring& orig) const
{
    if(orig.validate())
        return orig;

    if(isCharSetUTF8_)
        return StringToUTF8WithoutFail(orig, GetAssumeCharset());
    else
        return StringToUTF8WithoutFail(orig, currentCharSet_);
}
