/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_aac.cc
/// @brief Class for mpeg 4 audio decoding using FAAD2 source file.

#include <cstring>
#include <string>
#include <cstring>

#include <vector>

#include "../include/player_aac.h"
#include "../include/mainmanager.h"
#include "../include/scopeguard.h"

int FAAD::AACBuffer::adtsSampleRate_[] = {96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025, 8000, 7350, 0, 0, 0};

bool FAAD::AACBuffer::InitBuffer(const size_t& size)
{
    readSize_ = size;

    buffer_ = static_cast<unsigned char*>(std::malloc(size));
    if(buffer_ == NULL)
        return false;
    std::memset(buffer_, 0, size);

    const size_t bread = std::fread(buffer_, 1, size, inFile_);
    bytesIntoBuffer_ = bread;
    bytesConsumed_ = 0;
    fileOffset_ = 0;

    if(bread != size)
        atEof_ = true;

    return true;
}

void FAAD::AACBuffer::FillBuffer()
{
    if(bytesConsumed_ <= 0)
        return;

    if(bytesIntoBuffer_)
        memmove(static_cast<void*>(buffer_), static_cast<void*>(buffer_ + bytesConsumed_), bytesIntoBuffer_ * sizeof(unsigned char));

    if(!atEof_) {
        const int bread = std::fread(static_cast<void*>(buffer_ + bytesIntoBuffer_), 1, bytesConsumed_, inFile_);

        if(bread != bytesConsumed_)
            atEof_ = true;

        bytesIntoBuffer_ += bread;	

    }

    bytesConsumed_ = 0;

    if(bytesIntoBuffer_ > 3) {
        if(memcmp(buffer_, "TAG", 3) == 0)
            bytesIntoBuffer_ = 0;
    }

    if(bytesIntoBuffer_ > 11) {
        if(memcmp(buffer_, "LYRICSBEGIN", 11) == 0)
            bytesIntoBuffer_ = 0;
    }

    if(bytesIntoBuffer_ > 8) {
        if(memcmp(buffer_, "APETAGEX", 8) == 0)
            bytesIntoBuffer_ = 0;
    }
}

float FAAD::AACBuffer::GetLength()
{
    int tagSize = 0;
    if(memcmp(buffer_, "ID3", 3) != 0) {
        tagSize = (buffer_[6] << 21) | (buffer_[7] << 14) | (buffer_[8] << 7) |
            (buffer_[9] << 0);
        tagSize += 10;

        AdvanceBuffer(tagSize);
        FillBuffer();
    }

    float length = 0.0;

    if(buffer_[0] == 0xFF && (buffer_[1] & 0xF6) == 0xF0) {
        int frames = 0;
        int frameLength = 0;
        int sampleRate = 0;
        float framesPerSec = 0.0f;
        for(;; ++frames) {
            FillBuffer();
            if(bytesIntoBuffer_ > 7) {
                if(!(buffer_[0] == 0xFF && (buffer_[1] & 0xF6) == 0XF0))
                    break;

                if(frames == 0)
                    sampleRate = adtsSampleRate_[(buffer_[2]&0x3c)>>2];

                frameLength = ((static_cast<unsigned int>(buffer_[3]) & 0x3) << 11) | (static_cast<unsigned int>(buffer_[4]) << 3) | buffer_[5] >> 5;

                if(frameLength > bytesIntoBuffer_)
                    break;

                AdvanceBuffer(frameLength);

            } else
                break;
        }

        framesPerSec = sampleRate / 1024.0f;
        if(framesPerSec > 0.0f)
            length = frames / framesPerSec;
        
        std::fseek(inFile_, tagSize, SEEK_SET);
        const size_t bread = std::fread(buffer_, 1, readSize_, inFile_);
        if(bread != readSize_)
            atEof_ = true;
        else
            atEof_ = false;

        bytesIntoBuffer_ = bread;
        bytesConsumed_ = 0;
        fileOffset_ = tagSize; 

    } else if(std::memcmp(buffer_, "ADIF", 4) == 0) {
        const int skipSize = (buffer_[4] & 0x80) ? 9 : 0;
        const int bitRate =
            (static_cast<unsigned int>(buffer_[4 + skipSize] & 0x0F)<<19) |
            (static_cast<unsigned int>(buffer_[5 + skipSize])<<11) |
            (static_cast<unsigned int>(buffer_[6 + skipSize])<<3) |
            (static_cast<unsigned int>(buffer_[7 + skipSize]) & 0xE0);

        length = fileSize_;
        if(length != 0.0f && bitRate != 0)
            length = length * 8.0f / bitRate + 0.5f;
    }

    return length;
}

FAAD::AACPlayer::AACPlayer(MainManager& mainManager)
    : decodeState_(STOP),
      inverseSampleRate_(0.0),
      seekSeconds_(0),
      sampleRate_(0),
      maxChannels_(2),
      frameSize_(1024),
      mp4File_(false),
      channels_(0),
      inFile_(NULL),
      sampleBuffer_(NULL),
      mainManager_(mainManager)
{
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
}

FAAD::AACPlayer::~AACPlayer()
{
    Clean();
}

/// @brief Initialize player.
///
/// Open destination file and determine type.
bool FAAD::AACPlayer::Init(const std::string& filename)
{
    inFile_ = std::fopen(filename.c_str(), "rb");
    if(inFile_ == NULL)
        return false;

    mp4File_ = IsMP4File(inFile_);

    return true;
}

void FAAD::AACPlayer::Start()
{
    if(inFile_ != NULL) {
        decodeState_ = PLAY;

        if(mp4File_) {

            MP4Decode();

        } else {

            AACDecode();

        }

        if(decodeState_ == PLAY)
            mainManager_.AudioClose(true);
        else
            mainManager_.AudioClose(false);	

    }

    signalFinish_();
}

void FAAD::AACPlayer::Stop()
{
    decodeState_ = STOP;
}

void FAAD::AACPlayer::Seek(const unsigned long& seconds)
{
    decodeState_ = SEEK;
    seekSeconds_ = seconds;
}

/// @brief Decode MP4 type file.
/// @todo Need real testing. I have few MP4 encoded file, and works fine with them. But will work with others?
void FAAD::AACPlayer::MP4Decode()
{
    const std::auto_ptr<mp4ff_callback_t> mp4Cb(SetMP4ffCallback(inFile_));

    faacDecHandle decoder = faacDecOpen();
    ScopeGuard closeDecoder = MakeGuard(faacDecClose, decoder);

    SetDecoderConfig(decoder);

    mp4ff_t* const inFile = mp4ff_open_read(mp4Cb.get());
    if(inFile == NULL)
        return;
    ScopeGuard closeMP4ffFile = MakeGuard(mp4ff_close, inFile);

    int track = GetAACTrack(inFile);
    if(track < 0)
        return;

    unsigned char* buffer = NULL;
    unsigned int bufferSize = 0;
    mp4ff_get_decoder_config(inFile, track, &buffer, &bufferSize);

    if(faacDecInit2(decoder, buffer, bufferSize, &sampleRate_, &channels_) < 0)
        return;

    // Decoder was set to down matrix, but double check.
    if(static_cast<unsigned int>(channels_) > maxChannels_)
        return;

    mp4AudioSpecificConfig mp4ASC;
    if(buffer != NULL) {
        if(AudioSpecificConfig(buffer, bufferSize, &mp4ASC) >= 0) {
            if(mp4ASC.frameLengthFlag == 1)
                frameSize_ = 960;
            if(mp4ASC.sbr_present_flag == 1)
                frameSize_ *= 2;	
        }

        free(buffer);
        buffer = NULL;
        bufferSize = 0;
    }

    const long numSamples = mp4ff_num_samples(inFile, track);
    int rc = 0;
    faacDecFrameInfo frameInfo;

    if(!mainManager_.AudioOpenDevice(16, static_cast<unsigned int>(channels_), sampleRate_, mainManager_.audioFormatNativeEndian_))
        return;

    inverseSampleRate_ = 1.0 / sampleRate_;
    for(long sampleId = 0; sampleId < numSamples; ++sampleId) {
        if(decodeState_ == STOP)
            break;

        if(decodeState_ == SEEK) {

            sampleId = ConvertSecondToCount(seekSeconds_, mp4ASC.samplingFrequency, frameSize_);

            if(sampleId < 0)
                sampleId = 0;

            if(sampleId > numSamples)
                break;

            mainManager_.SetCurrentTime(seekSeconds_);

            decodeState_ = PLAY;

        }

        rc = mp4ff_read_sample(inFile, track, sampleId, &buffer, &bufferSize);
        if(rc == 0)
            return;

        sampleBuffer_ = faacDecDecode(decoder, &frameInfo, buffer, bufferSize);

        if(buffer != NULL) {
            free(buffer);
            buffer = NULL;
        }

        if(frameInfo.error == 0 && frameInfo.samples > 0) 
            mainManager_.AudioPlay(reinterpret_cast<double*>(sampleBuffer_), frameInfo.samples>>1);

    }
}

/// @brief Determine type.
bool FAAD::AACPlayer::IsMP4File(FILE* const inFile)
{
    unsigned char header[8];
    std::memset(header, 0, sizeof(header));
    std::fread(header, 1, 8, inFile);
    ScopeGuard rewindFile = MakeGuard(std::rewind, inFile);
    if(header[4] == 'f' && header[5] == 't' && header[6] == 'y' && header[7] == 'p')
        return true;

    return false;
}

int FAAD::AACPlayer::GetAACTrack(mp4ff_t* const inFile)
{
    const int totalTracks = mp4ff_total_tracks(inFile);
    int rc = 0;
    unsigned char* buffer = NULL;
    unsigned int bufferSize = 0;
    mp4AudioSpecificConfig mp4ASC;

    for(int i = 0; i < totalTracks; ++i) {
        mp4ff_get_decoder_config(inFile, i, &buffer, &bufferSize);

        if(buffer != NULL) {
            rc = AudioSpecificConfig(buffer, bufferSize, &mp4ASC);
            free(buffer);
            buffer = NULL;

            if(rc >= 0)
                return i;
        }
    }

    return -1;
}

/// @brief Decode AAC type file.
/// @todo Need real testing. Implement seek function.
void FAAD::AACPlayer::AACDecode()
{
    AACBuffer aacBuffer(inFile_);

    if(!aacBuffer.InitBuffer(FAAD_MIN_STREAMSIZE * maxChannels_))
        return;

    faacDecHandle decoder = faacDecOpen();
    ScopeGuard closeDecoder = MakeGuard(faacDecClose, decoder);

    SetDecoderConfig(decoder);

    aacBuffer.SkipInfo();

    aacBuffer.FillBuffer();
    if(!aacBuffer.InitDecoder(decoder, sampleRate_, channels_))
        return;

    // Decoder was set to down matrix, but double check.
    if(static_cast<unsigned int>(channels_) > maxChannels_)
        return;

    faacDecFrameInfo frameInfo;

    if(!mainManager_.AudioOpenDevice(16, static_cast<unsigned int>(channels_), sampleRate_, mainManager_.audioFormatNativeEndian_))
        return;

    inverseSampleRate_ = 1.0 / sampleRate_;
    do {
        if(decodeState_ == STOP)
            break;

        sampleBuffer_ = aacBuffer.Decode(decoder, frameInfo);

        aacBuffer.AdvanceBuffer(frameInfo.bytesconsumed);

        if(frameInfo.error == 0 && frameInfo.samples > 0) 
            mainManager_.AudioPlay(reinterpret_cast<double*>(sampleBuffer_), frameInfo.samples>>1);

        aacBuffer.FillBuffer();
        if(aacBuffer.IsEmpty())
            sampleBuffer_ = NULL;

    } while(sampleBuffer_ != NULL);

}

/// @brief Set mp4_callback_t.
std::auto_ptr<mp4ff_callback_t> FAAD::AACPlayer::SetMP4ffCallback(FILE* inFile)
{
    std::auto_ptr<mp4ff_callback_t> mp4Cb(new mp4ff_callback_t);
    mp4Cb->read = MP4ReadCallback;
    mp4Cb->seek = MP4SeekCallback;
    mp4Cb->user_data = inFile;

    return mp4Cb;
}

/// @brief Set decoder configuration.
/// @param decoder Decoder.
/// @param defObjectType Set object type.
/// @param outputFormat Library output formats. 
/// @param downMatrix Change multichannels to less one.
/// @param useOldADTSFormat Use old ADTS format.
/// @param dontUpSampleImplicitSBR SBR signalling.
void FAAD::AACPlayer::SetDecoderConfig(faacDecHandle& decoder, const unsigned char& defObjectType, const unsigned char& outputFormat, const unsigned char& downMatrix, const unsigned char& useOldADTSFormat, const unsigned char& dontUpSampleImplicitSBR)
{
    faacDecConfigurationPtr decoderConfig = faacDecGetCurrentConfiguration(decoder);
    decoderConfig->defObjectType = defObjectType;
    decoderConfig->outputFormat = outputFormat;
    decoderConfig->downMatrix = downMatrix;
    decoderConfig->useOldADTSFormat = useOldADTSFormat;
    decoderConfig->dontUpSampleImplicitSBR = dontUpSampleImplicitSBR;
    faacDecSetConfiguration(decoder, decoderConfig);
}

/// @brief Callback function used in mp4ff_callback_t.
uint32_t MP4ReadCallback(void* userData, void* buffer, uint32_t length)
{
    return std::fread(buffer, 1, length, static_cast<FILE*>(userData));
}

/// @brief Callback function used in mp4ff_callback_t.
uint32_t MP4SeekCallback(void* userData, uint64_t position)
{
    return std::fseek(static_cast<FILE*>(userData), position, SEEK_SET);
}

/// @brief Create AACPlayer object.
///
/// Used in PlayerFactory.
std::auto_ptr<Player> CreateAACPlayer(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new FAAD::AACPlayer(mainManager));
}
