/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// wma decoder came from bmp-wma plugin.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file player_wma.cc
/// @brief Class for play wma source file.

#include "../include/player_wma.h"
#include "../include/mainmanager.h"

/// @brief Constructor of WMAPlayer.
///
/// Connect dispatchers to functions.
FFMPEG::WMAPlayer::WMAPlayer(MainManager& mainManager)
    : decodeState_(STOP),
      seekSeconds_(0),
      bufSize_(4096),
      count_(bufSize_/2),
      sampleRate_(0),
      index_(0),
      validFile_(false),
      buf_(bufSize_),
      mainManager_(mainManager),
      avCodec_(NULL),
      avCodecContext_(NULL),
      avFormatContext_(NULL),
      outbuf_(NULL)
{
    signalFinish_.connect(sigc::mem_fun(mainManager_, &MainManager::PlayFinish));
}

FFMPEG::WMAPlayer::~WMAPlayer()
{
    Clean();
}

/// @brief Initialize WMAPlayer.
/// @param filename filename to play
/// @return Success of initialize
bool FFMPEG::WMAPlayer::Init(const std::string& filename)
{
    avcodec_init();
    avcodec_register_all();
    av_register_all();

    if(av_open_input_file(&avFormatContext_, filename.c_str(), NULL, 0, NULL) < 0) 
        return false;

    for(index_ = 0; index_ < avFormatContext_->nb_streams; ++index_)
        {
            avCodecContext_ = &avFormatContext_->streams[index_]->codec;
            if(avCodecContext_->codec_type == CODEC_TYPE_AUDIO)
                break;	
        }

    av_find_stream_info(avFormatContext_);

    avCodec_ = avcodec_find_decoder(avCodecContext_->codec_id);
    if(!avCodec_)
        return false;

    if(avcodec_open(avCodecContext_, avCodec_) < 0) 
        return false;

    outbuf_ = (uint8_t*)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);

    validFile_ = true;

    return true;
}

/// @brief Start player loop.
///
/// Decode file and send decoed samples to play function. Timer update and
/// audio open, close is also in this function.
void FFMPEG::WMAPlayer::Start()
{
    unsigned int channels = avCodecContext_->channels;
    if(!mainManager_.AudioCheckOpenDevice() || sampleRate_ != avCodecContext_->sample_rate) {
        sampleRate_ = avCodecContext_->sample_rate;
        mainManager_.AudioClose(true);
        if(!mainManager_.AudioOpenDevice(16, channels, sampleRate_, mainManager_.audioFormatNativeEndian_)) {

            signalFinish_();

            return;
        }
    }

    decodeState_ = PLAY;

    int size = 0;
    int len = 0;
    int out_size = 0;
    uint8_t* inbuf_ptr = NULL;

    for(;;) {

        {
            Glib::Mutex::Lock lock(avMutex_);

            if(decodeState_ == STOP)
                break;
            else if(decodeState_ == SEEK) {
                av_seek_frame(avFormatContext_, index_, seekSeconds_ * 1000000LL);
                mainManager_.SetCurrentTime(seekSeconds_);
                seekSeconds_ = 0;
                decodeState_ = PLAY;
            }
        }

        if(av_read_frame(avFormatContext_, &avPacket_) < 0)
            break;

        size = avPacket_.size;
        inbuf_ptr = avPacket_.data;

        if(size == 0)
            break;

        while(size > 0) {
            len = avcodec_decode_audio(avCodecContext_, 
                                       (short*)outbuf_,
                                       &out_size,
                                       inbuf_ptr,
                                       size);

            if(len < 0) 
                break;

            if(out_size <= 0)
                continue;	    
            else
                Play(out_size, channels);

            size -= len;
            inbuf_ptr += len;

            if(avPacket_.data)
                av_free_packet(&avPacket_);	
        }
    }

    if(decodeState_ == PLAY)
        mainManager_.AudioClose(true);
    else
        mainManager_.AudioClose(false);

    signalFinish_();
}

/// @brief Set decodeState to STOP.
///
/// Set decodeState to STOP. Detect STOP in Statr() function.
void FFMPEG::WMAPlayer::Stop()
{
    Glib::Mutex::Lock lock(avMutex_);

    decodeState_ = STOP;
}

/// @brief Set decodeState to SEEK and set seekSeconds to seconds..
///
/// Set decodeState to SEEK. Detect SEEK in Start() function.
/// @param seconds wanted time to seek in seconds.
void FFMPEG::WMAPlayer::Seek(const unsigned long& seconds)
{
    Glib::Mutex::Lock lock(avMutex_);

    decodeState_ = SEEK;

    seekSeconds_ = seconds;
}

void FFMPEG::WMAPlayer::ProcessStereo(short int* const tempSample, unsigned int& outbufSize, unsigned int& bufIndex, unsigned int& index)
{
    while(outbufSize >= bufSize_) {
        for(unsigned int i = 0; i < count_; ++i) {
            ShortIntSignalToDoubleSignal(tempSample, bufIndex, index);

            ShortIntSignalToDoubleSignal(tempSample, bufIndex, index);
        }

        mainManager_.AudioPlay(buf_, count_);
        index = 0;
        outbufSize -= bufSize_;
    }
}

void FFMPEG::WMAPlayer::ProcessMono(short int* const tempSample, unsigned int& outbufSize, unsigned int& bufIndex, unsigned int& index)
{
    while(outbufSize >= bufSize_) {
        for(unsigned int i = 0; i < count_; ++i) {
            ShortIntSignalToDoubleSignal(tempSample, bufIndex, index);

            NullSignal(bufIndex, index);
        }
        mainManager_.AudioPlay(buf_, count_);
        index = 0;
        outbufSize -= bufSize_;
    }
}

inline void FFMPEG::WMAPlayer::Play(const int& bufferSize, const unsigned int& channels)
{
    unsigned int index = 0;
    unsigned int bufIndex = 0;
    unsigned int outbufSize = bufferSize / 2;
    short int* tempSample = reinterpret_cast<short int*>(outbuf_);
    if(channels == 2)
        ProcessStereo(tempSample, outbufSize, bufIndex, index);
    else
        ProcessMono(tempSample, outbufSize, bufIndex, index);
}

/// @brief Create WMAPlayer object.
///
/// Used in PlayerFactory.
std::auto_ptr<Player> CreateWMAPlayer(MainManager& mainManager)
{
    return std::auto_ptr<Player>(new FFMPEG::WMAPlayer(mainManager));
}
