/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file playlist.cc
/// @brief Class for Playlist source file.

#include <iostream>
#include <fstream>

#include <vector>
#include <list>
#include <map>

#include <gdkmm.h>

#include <libintl.h>

#include "../include/playlist.h"
#include "../include/mainwin.h"
#include "../include/tag_mp3.h"
#include "../include/tag_ogg.h"
#include "../include/tag_wma.h"

#ifdef ENABLE_FAAD2
#include "../include/tag_aac.h"
#endif

#ifdef ENABLE_HTTP
#include "../include/connector.h"
#endif

#include "../include/fileinfodlg.h"
#include "../include/mainmanager.h"
#include "../include/trace.h"

const std::string Playlist::fileProtocol_("file://");
const unsigned int Playlist::mainWndPanelHeight_(170);
const unsigned int Playlist::eachRowHeight_(25);
// tagFactory_ will be destructed at end of program.
const std::auto_ptr<TagFactory> Playlist::tagFactory_(new TagFactory);
const std::auto_ptr<ListFactory> Playlist::listFactory_(new ListFactory);

/// @brief Constructor of Playlist.
/// @param mainWnd MainWnd
/// @param listId Identification string of list.
/// @param titleWidth Title column width.
/// @param loadlist Load list from listId file at first.
/// @todo Get MainWnd panel height from real window.
Playlist::Playlist(MainWnd& mainWnd, const std::string& listId, const unsigned int& titleWidth, const bool& loadlist)
    : listId_(listId),
      count_(0),
      tagProcessCount_(0),
      inTagParsing_(false),
      stopTagParsing_(false),
      treeModel_(Gtk::ListStore::create(columns_)),
      mainWnd_(mainWnd),
      tagParsingThread_(NULL)
{
    CalcTagProcessCount(mainWnd_.GetPrevHeight() - mainWndPanelHeight_);

    set_model(treeModel_);

    append_column(gettext("Number"), columns_.id_);
    append_column(gettext("Title"), columns_.title_);
    append_column(gettext("Time"), columns_.time_);

    CreatePopupMenuList();

    SetTitleColumnProperties(titleWidth);

    if(tagFactory_->Empty())
        RegisterTagFactory();

    if(listFactory_->Empty())
        RegisterListType();

    if(loadlist)
        LoadLastList();

    treeModel_->set_sort_func(columns_.filename_, sigc::mem_fun(*this, &Playlist::FilenameCompare));

    signalRowToFill_.connect(sigc::mem_fun(*this, &Playlist::FillRowFromQueue));

    std::list<Gtk::TargetEntry> targets;
    targets.push_back( Gtk::TargetEntry("text/plain") );
    drag_dest_set(targets);
    signal_drag_data_received().connect(sigc::mem_fun(*this, &Playlist::OnDragDataReceived));

    ForceUpdateToppage();
}

Playlist::~Playlist()
{
    StopThread();

    SaveList();
}

/// @brief Add file to list interface function.
///
/// If file is not playable and file is not list-type file, 
/// file will be rejected. Or call AddFile(const std::string& filePath) 
/// function to add a file that not parsed information.
/// @param filePath A filepath to add.
/// @param addListFile Load listfile if filepath is list-type file.
void Playlist::AddFile(const std::string& filePath, const bool& addListFile)
{
    if(mainWnd_.IsPlayable(filePath)) {

        AddFile(filePath);

    } else if(addListFile && IsListType(filePath)) {

        LoadList(filePath);

    }
}

/// @brief Add files to list.
/// @param files Files to add. 
/// @param addListFile Load listfile if filepath is list-type file.
void Playlist::AddFiles(const std::vector<std::string>& files, const bool& addListFile)
{
    typedef std::vector<std::string>::const_iterator iter;

    const iter end(files.end());
    for(iter eachFile(files.begin()); eachFile != end; ++eachFile) {

        AddFile(*eachFile, addListFile);

    }
}

/// @brief Add files under directory to playlist.
/// @param directoryPath Directory to add.
/// @param recursive Read directory recursively or not.
void Playlist::AddDirectory(const std::string& directoryPath, const bool& recursive)
{
    std::vector<std::string> filesToAdd;
    GetFilenameFromDirToVector(directoryPath, filesToAdd, recursive);

    AddFiles(filesToAdd, false);
}

#ifdef ENABLE_HTTP

/// @brief Add url to playlist interface function.
///
/// If url is INET file, just add it.
/// If url is some INET list file, get it and add it's contents.
/// Or add a url as stream.
/// @param Url to add.
void Playlist::AddUrl(const std::string& destUrl)
{
    if(IsListType(destUrl)) {

        std::string tmpFile(Glib::get_tmp_dir());
        tmpFile += "/sagua_tmplist.";
        tmpFile += GetExt(destUrl);

        GetListFromUrl(destUrl, tmpFile);

        LoadList(tmpFile);

        // check unlink error return.
        unlink(tmpFile.c_str());

    } else {

        const Glib::ustring convertedDestURL(mainWnd_.mainManager_.StringToUTF8(destUrl));

        AddFile(convertedDestURL, convertedDestURL, "--:--", 1);

    }
}

void Playlist::GetListFromUrl(const std::string& destUrl,
                              const std::string& listFile)
{
    // reduce these duplicated codes.
    bool redirect = false;
    unsigned int retry = 0;
    std::auto_ptr<Address> destAddress;
    std::string url(destUrl);

    do {

        destAddress = CreateAddress(url, mainWnd_.mainManager_.ConvertUrlToCurrentEncoding());

        if(destAddress.get() == NULL)
            return;

        ACE_INET_Addr srvr(destAddress->PortNumber(),
                           destAddress->RawHostName());

        ACE_Connector<FileGet, ACE_SOCK_CONNECTOR> connector;

        FileGet fileGet;
        fileGet.SetDestAddress(destAddress.get());
        fileGet.SetOutFile(listFile.c_str());
        FileGet* currentFileGet = &fileGet;
        
        if(connector.connect(currentFileGet, srvr) != -1) {

            ACE_Reactor::instance()->reset_reactor_event_loop();
            ACE_Reactor::instance()->run_reactor_event_loop();

        }

        if(redirect = fileGet.Redirect())
            url = fileGet.RedirectURL();

    } while(redirect && retry++ < 3);
}

#endif

/// @brief Get next position item's file path.
std::string Playlist::GetNext()
{
    if(pos_) {

        if(IsPositionLastRow()) {

            if(!mainWnd_.GetRepeatPlay()) {

                mainWnd_.SetListState(MainWnd::LISTSTATE_LAST);

                return std::string();
            }
        }

        return GetNextRowFilename();

    } else
        return GetCurrentPositionFilename();
}

/// @brief Get previous position item's file path.
std::string Playlist::GetPrev()
{
    if(pos_) {

        if(IsPositionFirstRow()) {

            if(mainWnd_.GetRepeatPlay())
                return GetLastRowFilename();
            else {

                mainWnd_.SetListState(MainWnd::LISTSTATE_FIRST);

                return std::string();
            }
        }

        return GetPrevRowFilename();

    } else
        return GetCurrentPositionFilename();
}

/// @brief Get item's file path from random position.
std::string Playlist::GetRandom()
{
    if(count_ == 0)
        return NullFilename();

    MoveNthPosition(GetRandomNumber());
    
    return GetCurrentPositionFilename();

}

/// @brief Force tag parsing from top page.
void Playlist::ForceUpdateToppage()
{
    const Gtk::TreeModel::iterator iter(treeModel_->children().begin());

    if(iter) {
        StopThread();
        inTagParsing_ = true;
        tagParsingThread_ = Glib::Thread::create( sigc::bind( sigc::mem_fun(*this, &Playlist::GetInfoFromTag), iter), true );
    }
}

/// @brief Force tag parsing from current view page.
void Playlist::ForceUpdatePage()
{
    while(Glib::MainContext::get_default()->iteration(false));

    OnVScrolled();
}

/// @brief Show current selected item's information.
void Playlist::ShowCurrentFileInfoDlg() const
{
    if(pos_)
        ShowFileInfoDlg( (*pos_)[columns_.filename_] );
}

/// @brief Callback function used when button is press.
///
/// If pressed button is 3 button, pop up menu.
bool Playlist::on_button_press_event(GdkEventButton* ev)
{
    const bool rtn(TreeView::on_button_press_event(ev));

    if( (ev->type == GDK_BUTTON_PRESS) && (ev->button == 3) ) {
        menuPopup_.popup(ev->button, ev->time);
    }

    return rtn;
}

/// @brief Callback function used when row is activated(double clicked or else ...)
void Playlist::on_row_activated(const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column)
{
    Gtk::TreeView::on_row_activated(path, column);

    const Gtk::TreeModel::iterator iter(treeModel_->get_iter(path));

    if(iter) {
        pos_ = iter;

        mainWnd_.SetPlayFilename(GetCurrentPositionFilename());

        mainWnd_.OnBtnPlay();
    }
}

/// @brief Callback function used when scroll adjustments is set.
///
/// This function called when TreeView regist hadj, vadj and unregist them.
/// OnVScrolled is connected vadj's signal_value_changed().
void Playlist::on_set_scroll_adjustments(Gtk::Adjustment* hadj, Gtk::Adjustment* vadj)
{
    Gtk::TreeView::on_set_scroll_adjustments(hadj, vadj);

    if(vadj) {
        vadj->signal_value_changed().connect( sigc::mem_fun(*this, &Playlist::OnVScrolled) );
    }

}

/// @brief Callback function used when size allocate.
///
/// This function called when widget is added to other container. 
/// tagProcessCount_ is calculated in here.
void Playlist::on_size_allocate(Gtk::Allocation& allocation)
{
    Gtk::TreeView::on_size_allocate(allocation);

    CalcTagProcessCount(allocation.get_height());
}

/// @brief Get file path from corressponding iterator.
///
/// Parse tag of file, move cursor and scroll to it.
std::string Playlist::GetCurrentPositionFilename()
{
    if(IsListEmpty())
        return NullFilename();

    if(!pos_)
        pos_ = treeModel_->children().begin();

    {
        Glib::Mutex::Lock lock(tagParsingMutex_);

        if( !(*pos_)[columns_.tagParsed_] ) { 
            ParseTag(pos_, false);
        }
    }

    mainWnd_.DisplayTitle(CurrentItemTitle());
    mainWnd_.DisplayTotalTime(CurrentItemTime());
    mainWnd_.SetSeekScaleRange(CurrentItemTimeSeconds());

    MoveCursorAndScroll(pos_);

    mainWnd_.SetListState(MainWnd::LISTSTATE_NORMAL);

    return std::string( (*pos_)[columns_.filename_] );

}

/// @brief Parse tag of file.
/// @param iter Iterator.
/// @param thread If function is called in thread, set true.
void Playlist::ParseTag(Gtk::TreeModel::iterator& iter, const bool& thread)
{
    const std::string filename((*iter)[columns_.filename_]);

    if(IsReadable(filename)) {
        std::auto_ptr<Tag> tagParser(tagFactory_->Create(filename));

        if(tagParser.get() != NULL) { 
            if(thread) {

                rowQueue_.push(new EachRow(iter, tagParser->GetTitle(), tagParser->GetTotalTime(), tagParser->GetTotalTimeSeconds()));

                signalRowToFill_();

            } else
                FillRow(iter, tagParser->GetTitle(), tagParser->GetTotalTime(), tagParser->GetTotalTimeSeconds());
        }
    }

}

/// @brief Fill row from rowQueue.
///
/// This function must be called in thread.
void Playlist::FillRowFromQueue()
{
    EachRow* processRow = NULL;

    {
        Glib::Mutex::Lock lock(tagParsingMutex_);

        if(rowQueue_.empty())
            return;

        processRow = rowQueue_.front();
        rowQueue_.pop();
    }

    FillRow(processRow->iter_, processRow->title_, processRow->time_, processRow->timeSeconds_);

    delete processRow;
}

/// @brief Parse tag loop.
/// @param iter Start iterator
void Playlist::GetInfoFromTag(Gtk::TreeModel::iterator& iter)
{
    unsigned int i = 0;

    while(!stopTagParsing_) {

        if(!iter || i > tagProcessCount_)
            break;

        {
            Glib::Mutex::Lock lock(tagParsingMutex_);

            if(!(*iter)[columns_.tagParsed_])
                ParseTag(iter, true);	

        }

        ++iter;
        ++i;

    }
}

/// @brief Show information dialog.
void Playlist::ShowFileInfoDlg(const std::string& filename) const
{
    if(IsReadable(filename)) {
        std::auto_ptr<Tag> tagParser(tagFactory_->Create(filename));

        if(tagParser.get() != NULL) { 
            std::map<std::string, Glib::ustring> fileInfoMap;

            fileInfoMap["title"] = tagParser->GetTitle();
            fileInfoMap["album"] = tagParser->GetAlbum();
            fileInfoMap["genre"] = tagParser->GetGenre();
            fileInfoMap["artist"] = tagParser->GetArtist();
            fileInfoMap["comment"] = tagParser->GetComment();
            fileInfoMap["year"] = tagParser->GetYear();
            fileInfoMap["track"] = tagParser->GetTrack();

            FileInfoDlg fileInfoDlg(fileInfoMap);
            fileInfoDlg.run();
        }
    }
}

/// @brief Callback function used when drag data received.
///
/// When drag data received, add them to list.
void Playlist::OnDragDataReceived(const Glib::RefPtr<Gdk::DragContext>& context, int x, int y, const Gtk::SelectionData& selectionData, guint info, guint time)
{
    if((selectionData.get_length() >=0 ) && (selectionData.get_format() == 8)) {

        AddDragedFiles(selectionData.get_data_as_string());

        ForceUpdateToppage();

    }

    context->drag_finish(false, false, time);
}

/// @brief Callback function used when remove menu is clicked.
void Playlist::OnMenuRemove()
{
    StopThread();

    if(Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection()) {
        if(Gtk::TreeModel::iterator iter = selection->get_selected()) {
            if(pos_ && pos_ == iter) {

                mainWnd_.EmptySelectedFilename();

                if(IsPositionLastRow())
                    pos_ = treeModel_->children().begin();
                else
                    ++pos_;

            }

            unsigned int count = (*iter)[columns_.id_];

            iter = treeModel_->erase(iter);
            MoveCursorAndScroll(iter);
            count_--;
            
            // Renewal id number. Do you have more effiecient idea?
            iter = treeModel_->children().begin();
            const Gtk::TreeModel::iterator end(treeModel_->children().end());
            while(iter != end) {
                if((*iter)[columns_.id_] > count) {
                    (*iter)[columns_.id_] = count;
                    ++count;
                }

                ++iter;
            }
        }
    }
}

/// @brief Callback function used when remove all menu is clicked.
void Playlist::OnMenuRemoveAll()
{
    StopThread();

    mainWnd_.EmptySelectedFilename();

    Gtk::TreeModel::iterator iter(treeModel_->children().begin());

    while(iter)
        iter = treeModel_->erase(iter);	

    count_ = 0;
    pos_ = treeModel_->children().begin();
}

/// @brief Callback function used when sort filename menu is clicked.
void Playlist::OnMenuSortFilename()
{
    treeModel_->set_sort_column(columns_.filename_, Gtk::SORT_ASCENDING);

    ForceUpdatePage();
}

/// @brief Callback function used when information menu is clicked.
void Playlist::OnMenuInfo()
{
    if(Glib::RefPtr<Gtk::TreeView::Selection> selection = get_selection()) {
        if(Gtk::TreeModel::iterator iter = selection->get_selected()) {

            ShowFileInfoDlg( (*iter)[columns_.filename_] );

        }
    }

}

/// @brief Callback function used when scrolled.
void Playlist::OnVScrolled()
{
    if(IsListEmpty())
        return;	

    StopThread();

    inTagParsing_ = true;
    tagParsingThread_ = Glib::Thread::create( sigc::bind( sigc::mem_fun(*this, &Playlist::GetInfoFromTag), GetTopPositionIterator()), true );
}

/// @brief Load list from indentification string file.
void Playlist::LoadLastList()
{
    Tracer::Trace trace("Playlist::LoadLastList");

    LoadList(mainWnd_.GetMainManager().GetProgramDirFilePath("/" + listId_ + ".m3u"));
}

/// @brief Load list from filename.
void Playlist::LoadList(const std::string& filename)
{
    Tracer::Trace trace("Playlist::LoadList");

    std::auto_ptr<ListFile> listFileProcessor(listFactory_->Create(filename));

    if(listFileProcessor.get() != NULL)
        listFileProcessor->Read(filename, *this);
}

/// @brief Save current list to identification string file.
void Playlist::SaveList()
{
    Tracer::Trace trace("Playlist::SaveList");

    if(!mainWnd_.GetMainManager().CheckProgramDir())
        return;

    M3UList listFile;
    listFile.Write(*this, mainWnd_.GetMainManager().GetProgramDirFilePath("/" + listId_ + ".m3u"));
}

/// @brief Stop thread.
void Playlist::StopThread()
{
    if(inTagParsing_) {
        stopTagParsing_ = true;
        tagParsingThread_->join();
        inTagParsing_ = false;	
        stopTagParsing_ = false;
    }
}

/// @brief Create popup menu list.
void Playlist::CreatePopupMenuList()
{
    Gtk::Menu_Helpers::MenuList& popupMenuList = menuPopup_.items();
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Add files"), sigc::mem_fun(mainWnd_, &MainWnd::OnMenuFileAdd) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Add a _directory"), sigc::mem_fun(mainWnd_, &MainWnd::OnMenuFileDirAdd) ));
#ifdef ENABLE_HTTP
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Add a URL"), sigc::mem_fun(mainWnd_, &MainWnd::OnMenuFileUrlAdd) ));
#endif
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Remove"), sigc::mem_fun(*this, &Playlist::OnMenuRemove) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("Re_move all"), sigc::mem_fun(*this, &Playlist::OnMenuRemoveAll) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Sort filename"), sigc::mem_fun(*this, &Playlist::OnMenuSortFilename) ));
    popupMenuList.push_back( Gtk::Menu_Helpers::SeparatorElem() );
    popupMenuList.push_back( Gtk::Menu_Helpers::MenuElem(gettext("_Information"), sigc::mem_fun(*this, &Playlist::OnMenuInfo) ));
}

/// @brief Register creator function to TagFactory.
void Playlist::RegisterTagFactory()
{
    tagFactory_->Register("mp1", &mainWnd_.mainManager_, &CreateId3Tag);
    tagFactory_->Register("mp2", &mainWnd_.mainManager_, &CreateId3Tag);
    tagFactory_->Register("mp3", &mainWnd_.mainManager_, &CreateId3Tag);
    tagFactory_->Register("ogg", &mainWnd_.mainManager_, &CreateOGGTag);
    tagFactory_->Register("wma", &mainWnd_.mainManager_, &CreateWMATag);

#ifdef ENABLE_FAAD2
    tagFactory_->Register("mp4", &mainWnd_.mainManager_, &CreateAACTag);
    tagFactory_->Register("m4a", &mainWnd_.mainManager_, &CreateAACTag);
    tagFactory_->Register("aac", &mainWnd_.mainManager_, &CreateAACTag);
#endif
}

void Playlist::RegisterListType()
{
    listFactory_->Register("m3u", &CreateM3UList);
    listFactory_->Register("pls", &CreatePLSList);
}

bool Playlist::IsListType(const std::string& filePath)
{
    return listFactory_->Exist(filePath);
}

/// @brief Add files from DND.
void Playlist::AddDragedFiles(const std::string& dragedString)
{
    std::vector<std::string> dragedItems;
    std::string::const_iterator itemBegin(dragedString.begin());
    std::string::const_iterator itemEnd;
    const std::string::const_iterator end(dragedString.end());

    for(;;) {

        itemBegin = find_if(itemBegin, end, std::not1(IsCharSpace()));

        if(itemBegin == end)
            break;

        itemEnd = find_if(itemBegin, end, IsCharSpace());

        if(itemBegin != itemEnd)
            dragedItems.push_back(std::string(itemBegin, itemEnd));

        itemBegin = itemEnd;

    }

    const std::vector<std::string>::const_iterator endItems(dragedItems.end());
    std::string eachItemName;

    for(std::vector<std::string>::const_iterator eachItem(dragedItems.begin());
        eachItem != endItems;
        ++eachItem) {

        if(eachItem->find(fileProtocol_, 0) != std::string::npos) {

            try {
                eachItemName = Glib::filename_from_uri(*eachItem);

                if(IsDirectory(eachItemName))
                    AddDirectory(eachItemName, true);
                else
                    AddFile(eachItemName, true);

            } catch(Glib::ConvertError& err) {

                // Got convert error!
            }
        }

#ifdef ENABLE_HTTP
        if(HasHttpProtocolString(*eachItem))
           AddUrl(*eachItem);
#endif
    }
}

/// @brief Set Liststate to empty and return null string.
inline std::string Playlist::NullFilename()
{
    mainWnd_.SetListState(MainWnd::LISTSTATE_EMPTY);
    
    return std::string();
}

void M3UList::Read(const std::string& filename, Playlist& output)
{
    std::ifstream listFile(filename.c_str());
    if(listFile) {

        std::string eachLine;
        const std::string dirname(Glib::path_get_dirname(filename) + '/');
        std::string tmp;

        while(getline(listFile, eachLine)) {
            if(IsComment(eachLine))
                continue;
            else {
                TrimLastSpaceCharacter(eachLine);
                // If line is not absolute path, assume that file in list file's dir.

                if(Glib::path_is_absolute(eachLine))
                    AddFile(eachLine, output);
#ifdef ENABLE_HTTP
                else if(HasHttpProtocolString(eachLine))
                    AddUrl(eachLine, output);
#endif
                else {
                    tmp = dirname;
                    tmp += eachLine;
                    AddFile(tmp, output);
                }
            }
        }
    }
}

void M3UList::Write(const Playlist& input, const std::string& filename)
{
    std::ofstream listFile(filename.c_str());
    if(listFile) {

        // m3u format... (is m3u a media queue format? anyway)
        //	#EXTM3U
        //	#EXTINF:song_length_in_seconds,song_title
        //	somefilepath.mp3
        listFile << "#EXTM3U" << std::endl;

#warning "hey, expection for not parsed file?"
        Gtk::TreeModel::iterator iter(GetFirstIter(input));

        unsigned int seconds = 0;
        Glib::ustring title;
        std::string filename;

        while(iter) {
            GetRawInformation(input, iter, seconds, title, filename);
            listFile << "#EXTINF:" << seconds << "," << title << std::endl 
                     << filename << std::endl;
            ++iter;
        }
    }
}

void PLSList::Read(const std::string& filename, Playlist& output)
{
    std::ifstream listFile(filename.c_str());
    if(listFile) {

        std::string eachLine;
        std::string::iterator iter;
        const std::string dirname(Glib::path_get_dirname(filename) + '/');
        std::string tmp;

        while(getline(listFile, eachLine)) {

            if(std::strncmp(eachLine.c_str(), "File", 4) == 0) {

                iter = find_if(eachLine.begin(), eachLine.end(), IsCharEqual());
                if(iter != eachLine.end()) {

                    eachLine = std::string(++iter, eachLine.end());
                    TrimLastSpaceCharacter(eachLine);

                    if(Glib::path_is_absolute(eachLine))
                        AddFile(eachLine, output);
#ifdef ENABLE_HTTP
                    else if(HasHttpProtocolString(eachLine))
                        AddUrl(eachLine, output);
#endif
                    else {

                        tmp = dirname;
                        tmp += eachLine;
                        AddFile(tmp, output);
                    }
                }
            }
        }
    }
}

void PLSList::Write(const Playlist& input, const std::string& filename)
{
    std::ofstream listFile(filename.c_str());
    if(listFile) {

        // pls format
        //
        // [playlist]
        // File1=some filename
        // Title1=some title
        // Length1=some length in seconds
        //			(negative number will be ignored, some stream)
        // File2=blah blah
        // Title2=blah blah
        // Length2=blah blah
        // 		...
        // NumberOfEntries=totle number of entry
        // Version=2
        //
        // All string case sensitive.
        listFile << "[playlist]" << std::endl;

#warning "hey, expection for not parsed file?"
        Gtk::TreeModel::iterator iter(GetFirstIter(input));

        unsigned int count = 0;
        unsigned int seconds = 0;
        Glib::ustring title;
        std::string filename;

        while(iter) {

            ++count;

            GetRawInformation(input, iter, seconds, title, filename);
            listFile << "File" << count << "=" << filename << std::endl;
            listFile << "Title" << count << "=" << title << std::endl;
            listFile << "Length" << count << "=" << seconds << std::endl;

            ++iter;

        }

        listFile << "NumberOfEntries=" << count << std::endl;
        listFile << "Version=2" << std::endl;
    }
}

std::auto_ptr<ListFile> CreateM3UList()
{
    return std::auto_ptr<ListFile>(new M3UList);
}

std::auto_ptr<ListFile> CreatePLSList()
{
    return std::auto_ptr<ListFile>(new PLSList);
}
