// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file sockethandler.cc
/// @brief Class for unix socket handler source file.

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/time.h>
#include <sys/select.h>

#include "../include/sockethandler.h"
#include "../include/mainmanager.h"

/// @brief Constructor of UnixSocketHandler.
/// @param arguments Command line arguments.
UnixSocketHandler::UnixSocketHandler(const std::vector<std::string>& arguments)
: arguments_(arguments),
    mainHeaderSize_(5 + UnixSocket::GetIntegerSize()),
    inThread_(false),
    stopThread_(false),
    acceptThread_(NULL),
    mainManager_(NULL)
{}

UnixSocketHandler::~UnixSocketHandler()
{
    if(inThread_) {
	stopThread_ = true;
	acceptThread_->join();
    }
}

/// @brief Send command line arguments to server.
void UnixSocketHandler::SendArguments()
{
    if(!arguments_.empty()) {

	unsigned int totalBytes = 0;
	GetTotalBytesToSend(totalBytes);

	SendHeader(totalBytes);

	SendData(totalBytes);

	unixSocket_.RecvSignal();

    }
}

/// @brief Set MainManager pointer.
void UnixSocketHandler::SetMainManager(MainManager* const mainManager)
{
    mainManager_ = mainManager;
    signalReceived_.connect(sigc::mem_fun(*mainManager_, &MainManager::SocketReceived));
}

/// @brief Send header.
///
/// Header : package version character(5 bytes) + data total bytes(in unsigned integer - probably 4 bytes)
inline void UnixSocketHandler::SendHeader(const unsigned int& totalBytes)
{
    char* const header = new char[mainHeaderSize_];
    strncpy(header, PACKAGE_VERSION, 5);
    memcpy(header + 5, reinterpret_cast<const char*>(&totalBytes), UnixSocket::GetIntegerSize());

    unixSocket_.Send(header, mainHeaderSize_);

    delete [] header;
}

/// @brief Send data.
///
/// Each data : size in bytes(in integer - probably 4 bytes) + real data
void UnixSocketHandler::SendData(const unsigned int& totalBytes)
{
    unsigned int thisBytes = 0;
    unsigned int processBytes = 0;
    char* const buf = new char[totalBytes];

    for(std::vector<std::string>::const_iterator iter = arguments_.begin();
	    iter != arguments_.end(); ++iter) {

	thisBytes = iter->size();
	memcpy(buf + processBytes, reinterpret_cast<char*>(&thisBytes), UnixSocket::GetIntegerSize());
	processBytes += UnixSocket::GetIntegerSize();

	memcpy(buf + processBytes, iter->c_str(), iter->size());
	processBytes += iter->size();
    }

    unixSocket_.Send(buf, totalBytes);

    delete [] buf;
}

/// @brief Loop for accept and communcation to client
void UnixSocketHandler::AcceptLoop()
{
    char buf[1024];
    fd_set master;
    fd_set readfds;
    struct sockaddr_un clientAddr;
    int clientLen = sizeof(clientAddr);
    int clientFd = -1;
    int fdmax = 0;
    int receivedBytes = -1;

    FD_ZERO(&master);
    FD_ZERO(&readfds);
    FD_SET(unixSocket_.GetSocket(), &master);

    fdmax = unixSocket_.GetSocket();

    struct timeval tv;

    while(!stopThread_) {
	tv.tv_sec = 0;
	tv.tv_usec = 1000;

	readfds = master;
	if(select(fdmax + 1, &readfds, NULL, NULL, &tv) == -1)
	    continue;

	for(int i = 0; i <= fdmax; ++i) {
	    if(FD_ISSET(i, &readfds)) {
		if(i == unixSocket_.GetSocket()) {
		    clientLen = sizeof(clientAddr);
		    clientFd = accept(unixSocket_.GetSocket(), reinterpret_cast<struct sockaddr*>(&clientAddr), reinterpret_cast<socklen_t*>(&clientLen));
		    if(clientFd == -1)
			continue;
		    else {
			FD_SET(clientFd, &master);
			if(clientFd > fdmax)
			    fdmax = clientFd;
		    }
		} else {
		    receivedBytes = -1;
		    receivedBytes = recv(i, buf, 1023, 0);
		    if(receivedBytes > 0)
			DataReceived(i, receivedBytes, buf, 1023);
		    else {
			close(i);
			FD_CLR(i, &master);
		    }
		}
	    }
	}
    }
}

/// @brief Process received data.
///
/// First, test data has valid header.
///
/// Header : maybe 9 bytes
///
///	    Package version charater(5 bytes) + Data total bytes(in integer - probably 4 bytes)
///
/// After, received all data to Data total bytes. And separate received data
/// to each element. Each element in received data has header for size.
///
/// Each element : size is variable
///
///	    Size in bytes(in integer - probably 4 bytes) + Real data 
///
/// @param fd Client socket
/// @param receivedBytes Received data size in first
/// @param buf Buffer to receive data
/// @param bufSize Buffer size
void UnixSocketHandler::DataReceived(const int& fd, const unsigned int& receivedBytes, char* const buf, const int& bufSize)
{
    // Header : package version character(5 bytes) + 
    //
    //	    data total bytes(in integer - probably 4 bytes) 

    // check version in received data
    const bool validData = receivedBytes >= mainHeaderSize_ && (strncmp(PACKAGE_VERSION, buf, 5) == 0);
    if(validData) {
	unsigned int totalBytes = 0;
	memcpy(&totalBytes, buf+5, UnixSocket::GetIntegerSize());

	char* const rcvData = new char[totalBytes];

	RecvEntireData(fd, buf, bufSize, receivedBytes, totalBytes, rcvData);

	// I got, thanks to client!
	UnixSocket::SendSignal(fd);

	ExtractReceivedData(rcvData, totalBytes);

	delete [] rcvData;
	if(mainManager_ != NULL)
	    signalReceived_();
    }
}

void UnixSocketHandler::RecvEntireData(const int& fd, char* const buf, const int& bufSize, const unsigned int& receivedBytes, unsigned int& totalBytes, char* const rcvData)
{
    int rcvBytes = -1;
    unsigned int curBytes = 0;
    unsigned int leftBytes = totalBytes;

    // If received not only header
    if(receivedBytes > mainHeaderSize_) {
	memcpy(rcvData, buf + mainHeaderSize_, receivedBytes - mainHeaderSize_);
	curBytes += receivedBytes - mainHeaderSize_;
	leftBytes -= receivedBytes - mainHeaderSize_;
    }

    // Received all data to expected total bytes
    while(curBytes < totalBytes) {

	rcvBytes = recv(fd, buf, bufSize, 0);
	// -1 is error and 0 is disconnection probably... 
	if(rcvBytes <= 0)
	    break;

	memcpy(rcvData + curBytes, buf, rcvBytes);

	curBytes += rcvBytes;
	leftBytes -= rcvBytes;  

    }

    // If cound not received entire data(client disconnection ?) 
    if(leftBytes > 0)
	totalBytes -= leftBytes;
}

void UnixSocketHandler::ExtractReceivedData(const char* const rcvData, const unsigned int& totalBytes)
{
    char* eachElem = NULL;
    unsigned int eachBytes = 0;
    unsigned int curBytes = 0;
    unsigned int leftBytes = totalBytes;

    Glib::Mutex::Lock lock(acceptMutex_);

    // This will isolate size part and real data part
    while(curBytes < totalBytes) {

	// Received data : size in bytes that not contain 
	//		terminator(in integer - probably 4 bytes) +
	//		
	//		real data

	if(leftBytes < UnixSocket::GetIntegerSize())
	    break;

	memcpy(&eachBytes, rcvData + curBytes, UnixSocket::GetIntegerSize());
	curBytes += UnixSocket::GetIntegerSize();
	leftBytes -= UnixSocket::GetIntegerSize();

	if(leftBytes < eachBytes)
	    break;

	// Need to allocate more 1 bytes for terminator
	eachElem = new char[eachBytes + 1];
	bzero(eachElem, eachBytes + 1);
	memcpy(eachElem, rcvData + curBytes, eachBytes);
	curBytes += eachBytes;
	leftBytes -= eachBytes;

	receivedData_.push_back(eachElem);

	delete [] eachElem;
    }
}

