/* -*- Mode: C++; tab-width: 4; indent-tabs-mod: nil; c-basic-offset: 4 -*- */
// Sagua is music player.
// Copyright (C) 2004 HG, Rhee <ez8tank@krpost.net>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/// @file tag_aac.cc
/// @brief Class for mpeg 4 tag processing source file.

#include <string>

#include "../include/tag_aac.h"
#include "../include/mainmanager.h"
#include "../include/scopeguard.h"
#include "../include/misc.h"

AACTag::AACTag(const std::string& filename, const MainManager* const mainManager)
    : filename_(filename),
      process_(false),
      mp4File_(false),
      inFile_(std::fopen(filename.c_str(), "rb")),
      mainManager_(mainManager)
{
    if(inFile_ != NULL)
        mp4File_ = FAAD::AACPlayer::IsMP4File(inFile_);
}

AACTag::~AACTag()
{
    if(inFile_ != NULL)
        std::fclose(inFile_);
}

/// @brief Get information from AAC type file.
///
/// @todo Need testing.
void AACTag::ProcessAAC()
{
    FAAD::AACBuffer aacBuffer(inFile_);

    if(!aacBuffer.InitBuffer(FAAD_MIN_STREAMSIZE))
        return;

    totalTimeSeconds_ = static_cast<unsigned long>(aacBuffer.GetLength());

    title_ = mainManager_->StringToUTF8(Glib::path_get_basename(filename_));

}

void AACTag::ProcessAll()
{
    if(inFile_ != NULL) {
        if(mp4File_) {
            MP4 mp4(filename_, inFile_, mainManager_);

            mp4.GetTotalTimeSeconds(totalTimeSeconds_);
            mp4.GetTitle(title_);
            mp4.GetArtist(artist_);
            mp4.GetAlbum(album_);
            mp4.GetTrack(track_);
            mp4.GetDate(year_);
            mp4.GetGenre(genre_);
            mp4.GetComment(comment_);
	    
        } else
            ProcessAAC();
    }
    
    process_ = true;
}

void AACTag::ProcessYear()
{
    if(!process_)
        ProcessAll();
}

void AACTag::ProcessTrack()
{
    if(!process_)
        ProcessAll();
}

void AACTag::ProcessTitle()
{
    if(title_.empty()) {
        if(!process_)
            ProcessAll();
        else
            title_ = mainManager_->StringToUTF8(Glib::path_get_basename(filename_));
    }
}

void AACTag::ProcessAlbum()
{
    if(!process_)
        ProcessAll();
}

void AACTag::ProcessGenre()
{
    if(!process_)
        ProcessAll();
}

void AACTag::ProcessArtist()
{
    if(!process_)
        ProcessAll();
}

void AACTag::ProcessComment()
{
    if(!process_)
        ProcessAll();
}

void AACTag::ProcessTotalTime()
{

    if(totalTime_.empty()) {

        if(!process_)
            ProcessAll();

        totalTime_ = ConvertSecondToTimeString(totalTimeSeconds_);	

    }
}

AACTag::MP4::MP4(const std::string& filename, FILE* const inFile, const MainManager* const mainManager)
    : filename_(filename),
      track_(-1),
      frameSize_(1024),
      valid_(true),
      decoder_(faacDecOpen()),
      mp4ffFile_(NULL),
      mp4Cb_(FAAD::AACPlayer::SetMP4ffCallback(inFile)),
      mainManager_(mainManager)
{
    valid_ = Init();

    mp4ffMetaCallbacks_["title"] = mp4ff_meta_get_title;
    mp4ffMetaCallbacks_["artist"] = mp4ff_meta_get_artist;
    mp4ffMetaCallbacks_["album"] = mp4ff_meta_get_album;
    mp4ffMetaCallbacks_["date"] = mp4ff_meta_get_date;
    mp4ffMetaCallbacks_["genre"] = mp4ff_meta_get_genre;
    mp4ffMetaCallbacks_["comment"] = mp4ff_meta_get_comment;
}

AACTag::MP4::~MP4()
{
    if(mp4ffFile_ != NULL)
        mp4ff_close(mp4ffFile_);

    faacDecClose(decoder_);
}

bool AACTag::MP4::Init()
{
    mp4ffFile_ = mp4ff_open_read(mp4Cb_.get());
    if(mp4ffFile_ == NULL)
        return false;

    track_ = FAAD::AACPlayer::GetAACTrack(mp4ffFile_);
    if(track_ < 0)
        return false;

    unsigned char* buffer = NULL;
    unsigned int bufferSize = 0;
    mp4ff_get_decoder_config(mp4ffFile_, track_, &buffer, &bufferSize);

    unsigned long sampleRate = 0;
    unsigned char channels = 0;
    if(faacDecInit2(decoder_, buffer, bufferSize, &sampleRate, &channels) < 0)
        return false;

    if(buffer != NULL) {
        if(AudioSpecificConfig(buffer, bufferSize, &mp4ASC_) >= 0) {
            if(mp4ASC_.frameLengthFlag == 1)
                frameSize_ = 960;
            if(mp4ASC_.sbr_present_flag == 1)
                frameSize_ *= 2;	
        }

        free(buffer);
        buffer = NULL;
        bufferSize = 0;
    }

    return true;
}

inline void AACTag::MP4::GetTotalTimeSeconds(unsigned long& totalTimeSeconds)
{
    if(valid_) {
        totalTimeSeconds = static_cast<unsigned long>(mp4ff_num_samples(mp4ffFile_, track_) * (frameSize_ - 1.0) / mp4ASC_.samplingFrequency);
    }
}

inline void AACTag::MP4::GetTitle(Glib::ustring& title)
{
    if(valid_) {

        GetMetaData("title", title);

        if(StringIsSpace(title))
            mainManager_->StringToUTF8(Glib::path_get_basename(filename_));
    }
}

inline void AACTag::MP4::GetArtist(Glib::ustring& artist)
{
    if(valid_)
        GetMetaData("artist", artist);
}

inline void AACTag::MP4::GetAlbum(Glib::ustring& album)
{
    if(valid_)
        GetMetaData("album", album);
}

inline void AACTag::MP4::GetTrack(Glib::ustring& track)
{
    if(valid_)
        GetMetaData("track", track);
}

inline void AACTag::MP4::GetDate(Glib::ustring& date)
{
    if(valid_)
        GetMetaData("date", date);
}

inline void AACTag::MP4::GetGenre(Glib::ustring& genre)
{
    if(valid_)
        GetMetaData("genre", genre);
}

inline void AACTag::MP4::GetComment(Glib::ustring& comment)
{
    if(valid_)
        GetMetaData("comment", comment);
}

void AACTag::MP4::GetMetaData(const std::string& key, Glib::ustring& output)
{
    Mp4ffMetaCallbacks::const_iterator iter = mp4ffMetaCallbacks_.find(key);
    if(iter != mp4ffMetaCallbacks_.end()) {

        char* value = NULL;
        if(iter->second(mp4ffFile_, &value)) {

            if(value != NULL) {
                output = value;
                free(value);

                mainManager_->StringToUTF8(output);
            }
        }
    }
}

/// @brief Create AACTag object.
///
/// Used in TagFactory.
std::auto_ptr<Tag> CreateAACTag(const std::string& filename, const MainManager* const mainManager)
{
    return std::auto_ptr<Tag>(new AACTag(filename, mainManager));
}
